# sage_setup: distribution = sagemath-categories
# sage.doctest: needs sage.modules
r"""
Example of a finite dimensional algebra with basis
"""
#*****************************************************************************
#  Copyright (C) 2008-2015 Franco Saliola <saliola@gmail.com>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#                  http://www.gnu.org/licenses/
#*****************************************************************************

from sage.misc.cachefunc import cached_method
from sage.categories.finite_dimensional_algebras_with_basis import FiniteDimensionalAlgebrasWithBasis
from sage.combinat.free_module import CombinatorialFreeModule


class KroneckerQuiverPathAlgebra(CombinatorialFreeModule):
    r"""
    An example of a finite dimensional algebra with basis: the path
    algebra of the Kronecker quiver.

    This class illustrates a minimal implementation of a finite
    dimensional algebra with basis. See
    :class:`sage.quivers.algebra.PathAlgebra` for a full-featured
    implementation of path algebras.
    """

    def __init__(self, base_ring):
        r"""
        EXAMPLES::

            sage: A = FiniteDimensionalAlgebrasWithBasis(QQ).example(); A
            An example of a finite dimensional algebra with basis:
            the path algebra of the Kronecker quiver
            (containing the arrows a:x->y and b:x->y) over Rational Field
            sage: TestSuite(A).run()
        """
        basis_keys = ['x', 'y', 'a', 'b']
        self._nonzero_products = {
            'xx':'x', 'xa':'a', 'xb':'b',
            'yy':'y', 'ay':'a', 'by':'b'
            }

        CombinatorialFreeModule.__init__(
            self, base_ring, basis_keys,
            category=FiniteDimensionalAlgebrasWithBasis(base_ring))

    def _repr_(self):
        r"""
        EXAMPLES::

            sage: FiniteDimensionalAlgebrasWithBasis(QQ).example() # indirect doctest
            An example of a finite dimensional algebra with basis:
            the path algebra of the Kronecker quiver
            (containing the arrows a:x->y and b:x->y) over Rational Field
        """
        return "An example of a finite dimensional algebra with basis: " \
            "the path algebra of the Kronecker quiver " \
            "(containing the arrows a:x->y and b:x->y) over %s " % (self.base_ring())

    def one(self):
        r"""
        Return the unit of this algebra.

        .. SEEALSO:: :meth:`AlgebrasWithBasis.ParentMethods.one_basis`

        EXAMPLES::

            sage: A = FiniteDimensionalAlgebrasWithBasis(QQ).example()
            sage: A.one()
            x + y
        """
        return self.sum_of_monomials(['x', 'y'])

    def product_on_basis(self, w1, w2):
        r"""
        Return the product of the two basis elements indexed by ``w1`` and ``w2``.

        .. SEEALSO:: :meth:`AlgebrasWithBasis.ParentMethods.product_on_basis`.

        EXAMPLES::

            sage: A = FiniteDimensionalAlgebrasWithBasis(QQ).example()

        Here is the multiplication table for the algebra::

            sage: matrix([[p*q for q in A.basis()] for p in A.basis()])
            [x 0 a b]
            [0 y 0 0]
            [0 a 0 0]
            [0 b 0 0]

        Here we take some products of linear combinations of basis elements::

            sage: x, y, a, b = A.basis()
            sage: a * (1-b)^2 * x
            0
            sage: x*a + b*y
            a + b
            sage: x*x
            x
            sage: x*y
            0
            sage: x*a*y
            a
        """
        if w1+w2 in self._nonzero_products:
            return self.monomial(self._nonzero_products[w1+w2])
        else:
            return self.zero()

    @cached_method
    def algebra_generators(self):
        r"""
        Return algebra generators for this algebra.

        .. SEEALSO:: :meth:`Algebras.ParentMethods.algebra_generators`.

        EXAMPLES::

            sage: A = FiniteDimensionalAlgebrasWithBasis(QQ).example(); A
            An example of a finite dimensional algebra with basis:
            the path algebra of the Kronecker quiver
            (containing the arrows a:x->y and b:x->y) over Rational Field
            sage: A.algebra_generators()
            Finite family {'x': x, 'y': y, 'a': a, 'b': b}
        """
        return self.basis()

    def _repr_term(self, p):
        r"""
        This method customizes the string representation of the basis
        element indexed by ``p``.

        In this example, we just return the string representation of
        ``p`` itself.

        EXAMPLES::

            sage: A = FiniteDimensionalAlgebrasWithBasis(QQ).example()
            sage: A.one()
            x + y
        """
        return str(p)


Example = KroneckerQuiverPathAlgebra



# sage_setup: distribution = sagemath-categories
r"""
Ideals of commutative rings

Sage provides functionality for computing with ideals. One can create
an ideal in any commutative or non-commutative ring `R` by giving a
list of generators, using the notation ``R.ideal([a,b,...])``. The case
of non-commutative rings is implemented in
:mod:`~sage.rings.noncommutative_ideals`.

A more convenient notation may be ``R*[a,b,...]`` or ``[a,b,...]*R``.
If `R` is non-commutative, the former creates a left and the latter
a right ideal, and ``R*[a,b,...]*R`` creates a two-sided ideal.
"""
# ****************************************************************************
#       Copyright (C) 2005 William Stein <wstein@gmail.com>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#
#    This code is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#    General Public License for more details.
#
#  The full text of the GPL is available at:
#
#                  https://www.gnu.org/licenses/
# ****************************************************************************

from types import GeneratorType

from sage.categories.rings import Rings
from sage.categories.fields import Fields
from sage.structure.element import MonoidElement
from sage.structure.richcmp import rich_to_bool, richcmp
from sage.structure.sequence import Sequence


# for efficiency
_Rings = Rings()
_Fields = Fields()


def Ideal(*args, **kwds):
    r"""
    Create the ideal in ring with given generators.

    There are some shorthand notations for creating an ideal, in
    addition to using the :func:`Ideal` function:

    - ``R.ideal(gens, coerce=True)``
    - ``gens*R``
    - ``R*gens``

    INPUT:

    - ``R`` -- a ring (optional; if not given, will try to infer it from ``gens``)

    - ``gens`` -- list of elements generating the ideal

    - ``coerce`` -- boolean (default: ``True``); whether ``gens`` need to be
      coerced into the ring

    OUTPUT: the ideal of ring generated by ``gens``

    EXAMPLES::

        sage: R.<x> = ZZ[]
        sage: I = R.ideal([4 + 3*x + x^2, 1 + x^2])
        sage: I
        Ideal (x^2 + 3*x + 4, x^2 + 1) of Univariate Polynomial Ring in x over Integer Ring
        sage: Ideal(R, [4 + 3*x + x^2, 1 + x^2])
        Ideal (x^2 + 3*x + 4, x^2 + 1) of Univariate Polynomial Ring in x over Integer Ring
        sage: Ideal((4 + 3*x + x^2, 1 + x^2))
        Ideal (x^2 + 3*x + 4, x^2 + 1) of Univariate Polynomial Ring in x over Integer Ring

    ::

        sage: ideal(x^2-2*x+1, x^2-1)
        Ideal (x^2 - 2*x + 1, x^2 - 1) of Univariate Polynomial Ring in x over Integer Ring
        sage: ideal([x^2-2*x+1, x^2-1])
        Ideal (x^2 - 2*x + 1, x^2 - 1) of Univariate Polynomial Ring in x over Integer Ring
        sage: l = [x^2-2*x+1, x^2-1]
        sage: ideal(f^2 for f in l)
        Ideal (x^4 - 4*x^3 + 6*x^2 - 4*x + 1, x^4 - 2*x^2 + 1) of
        Univariate Polynomial Ring in x over Integer Ring

    This example illustrates how Sage finds a common ambient ring for
    the ideal, even though 1 is in the integers (in this case).

    ::

        sage: R.<t> = ZZ['t']
        sage: i = ideal(1,t,t^2)
        sage: i
        Ideal (1, t, t^2) of Univariate Polynomial Ring in t over Integer Ring
        sage: ideal(1/2,t,t^2)
        Principal ideal (1) of Univariate Polynomial Ring in t over Rational Field

    This shows that the issues at :issue:`1104` are resolved::

        sage: Ideal(3, 5)
        Principal ideal (1) of Integer Ring
        sage: Ideal(ZZ, 3, 5)
        Principal ideal (1) of Integer Ring
        sage: Ideal(2, 4, 6)
        Principal ideal (2) of Integer Ring

    You have to provide enough information that Sage can figure out
    which ring to put the ideal in.

    ::

        sage: I = Ideal([])
        Traceback (most recent call last):
        ...
        ValueError: unable to determine which ring to embed the ideal in

        sage: I = Ideal()
        Traceback (most recent call last):
        ...
        ValueError: need at least one argument

    Note that some rings use different ideal implementations than the standard,
    even if they are PIDs.::

        sage: R.<x> = GF(5)[]
        sage: I = R * (x^2 + 3)
        sage: type(I)
        <class 'sage.rings.polynomial.ideal.Ideal_1poly_field'>

    You can also pass in a specific ideal type::

        sage: from sage.rings.ideal import Ideal_pid
        sage: I = Ideal(x^2+3,ideal_class=Ideal_pid)
        sage: type(I)
        <class 'sage.rings.ideal.Ideal_pid'>

    TESTS::

        sage: R.<x> = ZZ[]
        sage: I = R.ideal([4 + 3*x + x^2, 1 + x^2])
        sage: I == loads(dumps(I))
        True

    ::

        sage: I = Ideal(R, [4 + 3*x + x^2, 1 + x^2])
        sage: I == loads(dumps(I))
        True

    ::

        sage: I = Ideal((4 + 3*x + x^2, 1 + x^2))
        sage: I == loads(dumps(I))
        True

    This shows that the issue at :issue:`5477` is fixed::

        sage: R.<x> = QQ[]
        sage: I = R.ideal([x + x^2])
        sage: J = R.ideal([2*x + 2*x^2])
        sage: J
        Principal ideal (x^2 + x) of Univariate Polynomial Ring in x over Rational Field
        sage: S = R.quotient_ring(I)                                                    # needs sage.libs.pari
        sage: U = R.quotient_ring(J)                                                    # needs sage.libs.pari
        sage: I == J
        True
        sage: S == U                                                                    # needs sage.libs.pari
        True
    """
    if len(args) == 0:
        raise ValueError("need at least one argument")
    if len(args) == 1 and args[0] == []:
        raise ValueError("unable to determine which ring to embed the ideal in")

    first = args[0]

    inferred_field = False
    if first not in _Rings:
        if isinstance(first, Ideal_generic) and len(args) == 1:
            R = first.ring()
            gens = first.gens()
        else:
            if isinstance(first, (list, tuple, GeneratorType)) and len(args) == 1:
                gens = first
            else:
                gens = args
            gens = Sequence(gens)
            R = gens.universe()
            inferred_field = R in _Fields
    else:
        R = first
        gens = args[1:]

    if R not in _Rings.Commutative():
        raise TypeError("R must be a commutative ring")

    I = R.ideal(*gens, **kwds)

    if inferred_field and not isinstance(I, Ideal_fractional):  # trac 32320
        import warnings
        warnings.warn(f'Constructing an ideal in {R}, which is a field.'
                      ' Did you intend to take numerators first?'
                      ' This warning can be muted by passing the base ring to Ideal() explicitly.')

    return I


def is_Ideal(x):
    r"""
    Return ``True`` if object is an ideal of a ring.

    EXAMPLES:

    A simple example involving the ring of integers. Note
    that Sage does not interpret rings objects themselves as ideals.
    However, one can still explicitly construct these ideals::

        sage: from sage.rings.ideal import is_Ideal
        sage: R = ZZ
        sage: is_Ideal(R)
        doctest:warning...
        DeprecationWarning: The function is_Ideal is deprecated; use 'isinstance(..., Ideal_generic)' instead.
        See https://github.com/sagemath/sage/issues/38266 for details.
        False
        sage: 1*R; is_Ideal(1*R)
        Principal ideal (1) of Integer Ring
        True
        sage: 0*R; is_Ideal(0*R)
        Principal ideal (0) of Integer Ring
        True

    Sage recognizes ideals of polynomial rings as well::

        sage: R = PolynomialRing(QQ, 'x'); x = R.gen()
        sage: I = R.ideal(x^2 + 1); I
        Principal ideal (x^2 + 1) of Univariate Polynomial Ring in x over Rational Field
        sage: is_Ideal(I)
        True
        sage: is_Ideal((x^2 + 1)*R)
        True
    """
    from sage.misc.superseded import deprecation
    deprecation(38266,
                "The function is_Ideal is deprecated; "
                "use 'isinstance(..., Ideal_generic)' instead.")
    return isinstance(x, Ideal_generic)


class Ideal_generic(MonoidElement):
    """
    An ideal.

    See :func:`Ideal()`.
    """
    def __init__(self, ring, gens, coerce=True, **kwds):
        """
        Initialize this ideal.

        INPUT:

        - ``ring`` -- a ring

        - ``gens`` -- the generators for this ideal

        - ``coerce`` -- boolean (default: ``True``); if ``gens`` needs to be coerced
          into ``ring``

        EXAMPLES::

            sage: R.<x> = ZZ[]
            sage: R.ideal([4 + 3*x + x^2, 1 + x^2])
            Ideal (x^2 + 3*x + 4, x^2 + 1) of Univariate Polynomial Ring in x over Integer Ring
        """
        self.__ring = ring
        if not isinstance(gens, (list, tuple)):
            gens = [gens]
        if coerce:
            gens = [ring(x) for x in gens]

        gens = tuple(gens)
        if not gens:
            gens = (ring.zero(),)
        self.__gens = gens
        MonoidElement.__init__(self, ring.ideal_monoid())

    def _repr_short(self):
        """
        Represent the list of generators.

        EXAMPLES::

            sage: P.<a,b,c> = QQ[]
            sage: P*[a^2,a*b+c,c^3]
            Ideal (a^2, a*b + c, c^3) of Multivariate Polynomial Ring in a, b, c over Rational Field
            sage: (P*[a^2,a*b+c,c^3])._repr_short()
            '(a^2, a*b + c, c^3)'

        If the string representation of a generator contains a line break,
        the generators are not represented from left to right but from
        top to bottom. This is the case, e.g., for matrices::

            sage: MS = MatrixSpace(QQ, 2, 2)                                            # needs sage.modules
            sage: MS * [MS.1, 2]                                                        # needs sage.modules
            Left Ideal
            (
              [0 1]
              [0 0],
            <BLANKLINE>
              [2 0]
              [0 2]
            )
             of Full MatrixSpace of 2 by 2 dense matrices over Rational Field
        """
        L = []
        has_return = False
        for x in self.gens():
            s = repr(x)
            if '\n' in s:
                has_return = True
                s = s.replace('\n','\n  ')
            L.append(s)
        if has_return:
            return '\n(\n  %s\n)\n' % (',\n\n  '.join(L))
        return '(%s)' % (', '.join(L))

    def _repr_(self):
        """
        Return a string representation of ``self``.

        EXAMPLES::

            sage: P.<a,b,c> = QQ[]
            sage: P*[a^2,a*b+c,c^3] # indirect doctest
            Ideal (a^2, a*b + c, c^3) of Multivariate Polynomial Ring in a, b, c over Rational Field
        """
        return "Ideal %s of %s" % (self._repr_short(), self.ring())

    def random_element(self, *args, **kwds):
        """
        Return a random element in this ideal.

        EXAMPLES::

            sage: P.<a,b,c> = GF(5)[[]]
            sage: I = P.ideal([a^2, a*b + c, c^3])
            sage: I.random_element()  # random
            2*a^5*c + a^2*b*c^4 + ... + O(a, b, c)^13
        """
        return sum(self.__ring.random_element(*args, **kwds) * g for g in self.__gens)

    def _richcmp_(self, other, op):
        """
        Compare the generators of two ideals.

        INPUT:

        - ``other`` -- an ideal

        OUTPUT: boolean

        EXAMPLES::

            sage: R = ZZ; I = ZZ*2; J = ZZ*(-2)
            sage: I == J
            True
        """
        S = set(self.gens())
        T = set(other.gens())
        if S == T:
            return rich_to_bool(op, 0)
        return richcmp(self.gens(), other.gens(), op)

    def __contains__(self, x):
        """
        Check if ``x`` is in ``self``.

        EXAMPLES::

            sage: P.<a,b,c> = QQ[]
            sage: I = P * [a, b]
            sage: a + b in I                                                            # needs sage.libs.singular
            True
            sage: P2.<w,x,y,z> = QQ[]
            sage: x + 2*y + w*z in I
            False
        """
        try:
            return self._contains_(self.__ring(x))
        except TypeError:
            return False

    def _contains_(self, x):
        """
        Check if ``x``, which is assumed to be in the ambient ring, is in
        this ideal.

        .. TODO::

            Implement this method.

        EXAMPLES::

            sage: P.<a> = ZZ[]
            sage: I = P*[a]
            sage: I._contains_(a)
            Traceback (most recent call last):
            ...
            NotImplementedError

        Note that calling ``in`` does not call this method::

            sage: a in I
            True
        """
        raise NotImplementedError

    def __bool__(self):
        r"""
        Return ``True`` if this ideal is not `(0)`.

        TESTS::

            sage: I = ZZ.ideal(5)
            sage: bool(I)
            True

        ::

            sage: I = ZZ['x'].ideal(0)
            sage: bool(I)
            False

        ::

            sage: I = ZZ['x'].ideal(ZZ['x'].gen()^2)
            sage: bool(I)
            True

        ::

            sage: I = QQ['x', 'y'].ideal(0)
            sage: bool(I)
            False
        """
        for g in self.gens():
            if not g.is_zero():
                return True
        return False

    def base_ring(self):
        r"""
        Return the base ring of this ideal.

        EXAMPLES::

            sage: R = ZZ
            sage: I = 3*R; I
            Principal ideal (3) of Integer Ring
            sage: J = 2*I; J
            Principal ideal (6) of Integer Ring
            sage: I.base_ring(); J.base_ring()
            Integer Ring
            Integer Ring

        We construct an example of an ideal of a quotient ring::

            sage: R = PolynomialRing(QQ, 'x'); x = R.gen()
            sage: I = R.ideal(x^2 - 2)
            sage: I.base_ring()
            Rational Field

        And `p`-adic numbers::

            sage: R = Zp(7, prec=10); R                                                 # needs sage.rings.padics
            7-adic Ring with capped relative precision 10
            sage: I = 7*R; I                                                            # needs sage.rings.padics
            Principal ideal (7 + O(7^11)) of 7-adic Ring with capped relative precision 10
            sage: I.base_ring()                                                         # needs sage.rings.padics
            7-adic Ring with capped relative precision 10
        """
        return self.ring().base_ring()

    def apply_morphism(self, phi):
        r"""
        Apply the morphism ``phi`` to every element of this ideal.
        Returns an ideal in the domain of ``phi``.

        EXAMPLES::

            sage: # needs sage.rings.real_mpfr
            sage: psi = CC['x'].hom([-CC['x'].0])
            sage: J = ideal([CC['x'].0 + 1]); J
            Principal ideal (x + 1.00000000000000) of Univariate Polynomial Ring in x
             over Complex Field with 53 bits of precision
            sage: psi(J)
            Principal ideal (x - 1.00000000000000) of Univariate Polynomial Ring in x
             over Complex Field with 53 bits of precision
            sage: J.apply_morphism(psi)
            Principal ideal (x - 1.00000000000000) of Univariate Polynomial Ring in x
             over Complex Field with 53 bits of precision

        ::

            sage: psi = ZZ['x'].hom([-ZZ['x'].0])
            sage: J = ideal([ZZ['x'].0, 2]); J
            Ideal (x, 2) of Univariate Polynomial Ring in x over Integer Ring
            sage: psi(J)
            Ideal (-x, 2) of Univariate Polynomial Ring in x over Integer Ring
            sage: J.apply_morphism(psi)
            Ideal (-x, 2) of Univariate Polynomial Ring in x over Integer Ring

        TESTS::

            sage: # needs sage.rings.number_fields
            sage: x = polygen(ZZ)
            sage: K.<a> = NumberField(x^2 + 1)
            sage: A = K.ideal(a)
            sage: taus = K.embeddings(K)
            sage: A.apply_morphism(taus[0]) # identity
            Fractional ideal (a)
            sage: A.apply_morphism(taus[1]) # complex conjugation
            Fractional ideal (-a)
            sage: A.apply_morphism(taus[0]) == A.apply_morphism(taus[1])
            True

        ::

            sage: # needs sage.rings.number_fields
            sage: K.<a> = NumberField(x^2 + 5)
            sage: B = K.ideal([2, a + 1]); B
            Fractional ideal (2, a + 1)
            sage: taus = K.embeddings(K)
            sage: B.apply_morphism(taus[0]) # identity
            Fractional ideal (2, a + 1)

        Since 2 is totally ramified, complex conjugation fixes it::

            sage: B.apply_morphism(taus[1])  # complex conjugation                      # needs sage.rings.number_fields
            Fractional ideal (2, a + 1)
            sage: taus[1](B)                                                            # needs sage.rings.number_fields
            Fractional ideal (2, a + 1)
        """
        from sage.categories.morphism import Morphism
        if not isinstance(phi, Morphism):
            raise TypeError("phi must be a morphism")
        # delegate: morphisms know how to apply themselves to ideals
        return phi(self)

    def _latex_(self):
        r"""
        Return a latex representation of ``self``.

        EXAMPLES::

            sage: latex(3*ZZ) # indirect doctest
            \left(3\right)\Bold{Z}
        """
        import sage.misc.latex as latex
        return '\\left(%s\\right)%s' % (", ".join(latex.latex(g)
                                                  for g in self.gens()),
                                        latex.latex(self.ring()))

    def ring(self):
        """
        Return the ring containing this ideal.

        EXAMPLES::

            sage: R = ZZ
            sage: I = 3*R; I
            Principal ideal (3) of Integer Ring
            sage: J = 2*I; J
            Principal ideal (6) of Integer Ring
            sage: I.ring(); J.ring()
            Integer Ring
            Integer Ring

        Note that ``self.ring()`` is different from
        ``self.base_ring()``

        ::

            sage: R = PolynomialRing(QQ, 'x'); x = R.gen()
            sage: I = R.ideal(x^2 - 2)
            sage: I.base_ring()
            Rational Field
            sage: I.ring()
            Univariate Polynomial Ring in x over Rational Field

        Another example using polynomial rings::

            sage: R = PolynomialRing(QQ, 'x'); x = R.gen()
            sage: I = R.ideal(x^2 - 3)
            sage: I.ring()
            Univariate Polynomial Ring in x over Rational Field
            sage: Rbar = R.quotient(I, names='a')                                       # needs sage.libs.pari
            sage: S = PolynomialRing(Rbar, 'y'); y = Rbar.gen(); S                      # needs sage.libs.pari
            Univariate Polynomial Ring in y over
             Univariate Quotient Polynomial Ring in a over Rational Field with modulus x^2 - 3
            sage: J = S.ideal(y^2 + 1)                                                  # needs sage.libs.pari
            sage: J.ring()                                                              # needs sage.libs.pari
            Univariate Polynomial Ring in y over
             Univariate Quotient Polynomial Ring in a over Rational Field with modulus x^2 - 3
        """
        return self.__ring

    def reduce(self, f):
        r"""
        Return the reduction of the element of `f` modulo ``self``.

        This is an element of `R` that is equivalent modulo `I` to `f` where
        `I` is ``self``.

        EXAMPLES::

            sage: ZZ.ideal(5).reduce(17)
            2
            sage: parent(ZZ.ideal(5).reduce(17))
            Integer Ring
        """
        return f       # default

    def gens(self):  # -> tuple | PolynomialSequence
        """
        Return a set of generators / a basis of ``self``.

        This is the set of generators provided during creation of this ideal.

        EXAMPLES::

            sage: P.<x,y> = PolynomialRing(QQ,2)
            sage: I = Ideal([x,y+1]); I
            Ideal (x, y + 1) of Multivariate Polynomial Ring in x, y over Rational Field
            sage: I.gens()
            [x, y + 1]

        ::

            sage: ZZ.ideal(5,10).gens()
            (5,)
        """
        return self.__gens

    def gen(self, i):
        """
        Return the ``i``-th generator in the current basis of this ideal.

        EXAMPLES::

            sage: P.<x,y> = PolynomialRing(QQ,2)
            sage: I = Ideal([x,y+1]); I
            Ideal (x, y + 1) of Multivariate Polynomial Ring in x, y over Rational Field
            sage: I.gen(1)
            y + 1

            sage: ZZ.ideal(5,10).gen()
            5
        """
        return self.__gens[i]

    def ngens(self):
        """
        Return the number of generators in the basis.

        EXAMPLES::

            sage: P.<x,y> = PolynomialRing(QQ,2)
            sage: I = Ideal([x,y+1]); I
            Ideal (x, y + 1) of Multivariate Polynomial Ring in x, y over Rational Field
            sage: I.ngens()
            2

            sage: ZZ.ideal(5,10).ngens()
            1
        """
        return len(self.__gens)

    def gens_reduced(self):
        r"""
        Same as :meth:`gens()` for this ideal, since there is currently no
        special ``gens_reduced`` algorithm implemented for this ring.

        This method is provided so that ideals in `\ZZ` have the method
        ``gens_reduced()``, just like ideals of number fields.

        EXAMPLES::

            sage: ZZ.ideal(5).gens_reduced()
            (5,)
        """
        return self.gens()

    def is_maximal(self):
        r"""
        Return ``True`` if the ideal is maximal in the ring containing the
        ideal.

        .. TODO::

            This is not implemented for many rings.  Implement it!

        EXAMPLES::

            sage: R = ZZ
            sage: I = R.ideal(7)
            sage: I.is_maximal()
            True
            sage: R.ideal(16).is_maximal()
            False
            sage: S = Integers(8)
            sage: S.ideal(0).is_maximal()
            False
            sage: S.ideal(2).is_maximal()
            True
            sage: S.ideal(4).is_maximal()
            False
        """
        from sage.rings.integer_ring import ZZ
        R = self.ring()
        if hasattr(R, 'cover_ring') and R.cover_ring() is ZZ:
            # The following test only works for quotients of Z/nZ: for
            # many other rings in Sage, testing whether R/I is a field
            # is done by testing whether I is maximal, so this would
            # result in a loop.
            return R.quotient(self).is_field()
        kd = R.krull_dimension()
        if kd == 0 or (kd == 1 and R.is_integral_domain()):
            # For rings of Krull dimension 0, or for integral domains of
            # Krull dimension 1, every nontrivial prime ideal is maximal.
            return self.is_prime()
        else:
            raise NotImplementedError

    def is_primary(self, P=None):
        r"""
        Return ``True`` if this ideal is primary (or `P`-primary, if
        a prime ideal `P` is specified).

        Recall that an ideal `I` is primary if and only if `I` has a
        unique associated prime (see page 52 in [AM1969]_).  If this
        prime is `P`, then `I` is said to be `P`-primary.

        INPUT:

        - ``P`` -- (default: ``None``) a prime ideal in the same ring

        EXAMPLES::

            sage: R.<x, y> = QQ[]
            sage: I = R.ideal([x^2, x*y])
            sage: I.is_primary()                                                        # needs sage.libs.singular
            False
            sage: J = I.primary_decomposition()[1]; J                                   # needs sage.libs.singular
            Ideal (y, x^2) of Multivariate Polynomial Ring in x, y over Rational Field
            sage: J.is_primary()                                                        # needs sage.libs.singular
            True
            sage: J.is_prime()                                                          # needs sage.libs.singular
            False

        Some examples from the Macaulay2 documentation::

            sage: # needs sage.rings.finite_rings
            sage: R.<x, y, z> = GF(101)[]
            sage: I = R.ideal([y^6])
            sage: I.is_primary()                                                        # needs sage.libs.singular
            True
            sage: I.is_primary(R.ideal([y]))                                            # needs sage.libs.singular
            True
            sage: I = R.ideal([x^4, y^7])
            sage: I.is_primary()                                                        # needs sage.libs.singular
            True
            sage: I = R.ideal([x*y, y^2])
            sage: I.is_primary()                                                        # needs sage.libs.singular
            False

        .. NOTE::

            This uses the list of associated primes.
        """
        try:
            ass = self.associated_primes()
        except (NotImplementedError, ValueError):
            raise NotImplementedError
        if P is None:
            return (len(ass) == 1)
        else:
            return (len(ass) == 1) and (ass[0] == P)

    def primary_decomposition(self):
        r"""
        Return a decomposition of this ideal into primary ideals.

        EXAMPLES::

            sage: R = ZZ['x']
            sage: I = R.ideal(7)
            sage: I.primary_decomposition()
            Traceback (most recent call last):
            ...
            NotImplementedError
        """
        raise NotImplementedError

    def is_prime(self):
        r"""
        Return ``True`` if this ideal is prime.

        EXAMPLES::

            sage: R.<x, y> = QQ[]
            sage: I = R.ideal([x, y])
            sage: I.is_prime()        # a maximal ideal                                 # needs sage.libs.singular
            True
            sage: I = R.ideal([x^2 - y])
            sage: I.is_prime()        # a non-maximal prime ideal                       # needs sage.libs.singular
            True
            sage: I = R.ideal([x^2, y])
            sage: I.is_prime()        # a non-prime primary ideal                       # needs sage.libs.singular
            False
            sage: I = R.ideal([x^2, x*y])
            sage: I.is_prime()        # a non-prime non-primary ideal                   # needs sage.libs.singular
            False

            sage: S = Integers(8)
            sage: S.ideal(0).is_prime()
            False
            sage: S.ideal(2).is_prime()
            True
            sage: S.ideal(4).is_prime()
            False

        Note that this method is not implemented for all rings where it
        could be::

            sage: R.<x> = ZZ[]
            sage: I = R.ideal(7)
            sage: I.is_prime()        # when implemented, should be True
            Traceback (most recent call last):
            ...
            NotImplementedError

        .. NOTE::

            For general rings, uses the list of associated primes.
        """
        from sage.rings.integer_ring import ZZ
        R = self.ring()
        if hasattr(R, 'cover_ring') and R.cover_ring() is ZZ and R.is_finite():
            # For quotient rings of ZZ, prime is the same as maximal.
            return self.is_maximal()
        try:
            ass = self.associated_primes()
        except (NotImplementedError, ValueError):
            raise NotImplementedError
        if len(ass) != 1:
            return False
        if self == ass[0]:
            return True
        else:
            return False

    def associated_primes(self):
        r"""
        Return the list of associated prime ideals of this ideal.

        EXAMPLES::

            sage: R = ZZ['x']
            sage: I = R.ideal(7)
            sage: I.associated_primes()
            Traceback (most recent call last):
            ...
            NotImplementedError
        """
        raise NotImplementedError

    def minimal_associated_primes(self):
        r"""
        Return the list of minimal associated prime ideals of this ideal.

        EXAMPLES::

            sage: R = ZZ['x']
            sage: I = R.ideal(7)
            sage: I.minimal_associated_primes()
            Traceback (most recent call last):
            ...
            NotImplementedError
        """
        raise NotImplementedError

    def embedded_primes(self):
        r"""
        Return the list of embedded primes of this ideal.

        EXAMPLES::

            sage: R.<x, y> = QQ[]
            sage: I = R.ideal(x^2, x*y)
            sage: I.embedded_primes()                                                   # needs sage.libs.singular
            [Ideal (y, x) of Multivariate Polynomial Ring in x, y over Rational Field]
        """
        # by definition, embedded primes are associated primes that
        # are not minimal (under inclusion)
        ass = self.associated_primes()
        min_ass = self.minimal_associated_primes()
        emb = []
        for p in ass:
            try:
                min_ass.index(p)
            except ValueError:
                emb.append(p)
        emb.sort()
        return emb

    def is_principal(self):
        r"""
        Return ``True`` if the ideal is principal in the ring containing the
        ideal.

        .. TODO::

            Code is naive. Only keeps track of ideal generators as set
            during initialization of the ideal. (Can the base ring change? See
            example below.)

        EXAMPLES::

            sage: R.<x> = ZZ[]
            sage: I = R.ideal(2, x)
            sage: I.is_principal()
            Traceback (most recent call last):
            ...
            NotImplementedError
            sage: J = R.base_extend(QQ).ideal(2, x)
            sage: J.is_principal()
            True
        """
        if len(self.gens()) <= 1:
            return True
        raise NotImplementedError

    def is_trivial(self):
        r"""
        Return ``True`` if this ideal is `(0)` or `(1)`.

        TESTS::

            sage: I = ZZ.ideal(5)
            sage: I.is_trivial()
            False

        ::

            sage: I = ZZ['x'].ideal(-1)
            sage: I.is_trivial()
            True

        ::

            sage: I = ZZ['x'].ideal(ZZ['x'].gen()^2)
            sage: I.is_trivial()
            False

        ::

            sage: I = QQ['x', 'y'].ideal(-5)
            sage: I.is_trivial()                                                        # needs sage.libs.singular
            True

        ::

            sage: I = CC['x'].ideal(0)                                                  # needs sage.rings.real_mpfr
            sage: I.is_trivial()                                                        # needs sage.rings.real_mpfr
            True

        This test addresses issue :issue:`20514`::

            sage: R = QQ['x', 'y']
            sage: I = R.ideal(R.gens())
            sage: I.is_trivial()                                                        # needs sage.libs.singular
            False
        """
        return self.is_zero() or self == self.ring().unit_ideal()

    def category(self):
        """
        Return the category of this ideal.

        .. NOTE::

            category is dependent on the ring of the ideal.

        EXAMPLES::

            sage: P.<x> = ZZ[]
            sage: I = ZZ.ideal(7)
            sage: J = P.ideal(7,x)
            sage: K = P.ideal(7)
            sage: I.category()
            Category of ring ideals in Integer Ring
            sage: J.category()
            Category of ring ideals in Univariate Polynomial Ring in x
            over Integer Ring
            sage: K.category()
            Category of ring ideals in Univariate Polynomial Ring in x
            over Integer Ring
        """
        import sage.categories.all
        return sage.categories.all.Ideals(self.__ring)

    def __add__(self, other):
        """
        Add ``self`` on the left to ``other``.

        This makes sure that ``other`` and ``self`` are in the same rings.

        EXAMPLES::

            sage: P.<x,y,z> = QQ[]
            sage: I = [x + y]*P
            sage: I + [y + z]
            Ideal (x + y, y + z) of Multivariate Polynomial Ring in x, y, z over Rational Field
        """
        if not isinstance(other, Ideal_generic):
            other = self.ring().ideal(other)
        return self.ring().ideal(self.gens() + other.gens())

    def __radd__(self, other):
        """
        Add ``self`` on the right to ``other``.

        This makes sure that ``other`` and ``self`` are in the same rings.

        EXAMPLES::

            sage: P.<x,y,z> = QQ[]
            sage: I = [x + y]*P
            sage: [y + z] + I
            Ideal (x + y, y + z) of Multivariate Polynomial Ring in x, y, z over Rational Field
        """
        if not isinstance(other, Ideal_generic):
            other = self.ring().ideal(other)
        return self.ring().ideal(self.gens() + other.gens())

    def __mul__(self, other):
        """
        This method just makes sure that ``self`` and ``other`` are ideals in the
        same ring and then calls :meth:`_mul_`. If you want to change the
        behaviour of ideal multiplication in a subclass of
        :class:`Ideal_generic` please overwrite :meth:`_mul_` and not
        :meth:`__mul__`.

        EXAMPLES::

            sage: P.<x,y,z> = QQ[]
            sage: I = [x*y + y*z, x^2 + x*y - y*x - y^2] * P
            sage: I * 2    # indirect doctest
            Ideal (2*x*y + 2*y*z, 2*x^2 - 2*y^2) of Multivariate Polynomial Ring in x, y, z over Rational Field
        """
        if not isinstance(other, Ideal_generic):
            try:
                if self.ring().has_coerce_map_from(other):
                    return self
            except (TypeError,ArithmeticError,ValueError):
                pass
            other = self.ring().ideal(other)
        return self._mul_(other)

    def _mul_(self, other):
        """
        This is a very general implementation of Ideal multiplication.

        This method assumes that ``self`` and ``other`` are Ideals of the
        same ring.

        The number of generators of ``self * other`` will be
        ``self.ngens() * other.ngens()``. So if used repeatedly this method
        will create an ideal with a uselessly large amount of generators.
        Therefore it is advisable to overwrite this method with a method that
        takes advantage of the structure of the ring you are working in.

        Example::

            sage: P.<x,y,z> = QQ[]
            sage: I=P.ideal([x*y, x*z, x^2])
            sage: J=P.ideal([x^2, x*y])
            sage: I._mul_(J)
            Ideal (x^3*y, x^2*y^2, x^3*z, x^2*y*z, x^4, x^3*y) of Multivariate Polynomial Ring in x, y, z over Rational Field
        """
        return self.ring().ideal([z for z in [x*y for x in self.gens() for y in other.gens()] if z])

    def __rmul__(self, other):
        """
        Multiply ``self`` on the right with ``other``.

        EXAMPLES::

            sage: P.<x,y,z> = QQ[]
            sage: I = [x*y+y*z,x^2+x*y-y*x-y^2]*P
            sage: [2]*I    # indirect doctest
            Ideal (2*x*y + 2*y*z, 2*x^2 - 2*y^2) of Multivariate Polynomial Ring in x, y, z over Rational Field
        """
        if not isinstance(other, Ideal_generic):
            try:
                if self.ring().has_coerce_map_from(other):
                    return self
            except (TypeError,ArithmeticError,ValueError):
                pass
            other = self.ring().ideal(other)
        return self.ring().ideal([z for z in [y*x for x in self.gens() for y in other.gens()] if z])

    def norm(self):
        """
        Return the norm of this ideal.

        In the general case, this is just the ideal itself, since the ring it
        lies in can't be implicitly assumed to be an extension of anything.

        We include this function for compatibility with cases such as ideals in
        number fields.

        EXAMPLES::

            sage: R.<t> = GF(8, names='a')[]                                            # needs sage.rings.finite_rings
            sage: I = R.ideal(t^4 + t + 1)                                              # needs sage.rings.finite_rings
            sage: I.norm()                                                              # needs sage.rings.finite_rings
            Principal ideal (t^4 + t + 1) of Univariate Polynomial Ring in t
             over Finite Field in a of size 2^3
        """
        return self

    def absolute_norm(self):
        """
        Return the absolute norm of this ideal.

        In the general case, this is just the ideal itself, since the ring it
        lies in can't be implicitly assumed to be an extension of anything.

        We include this function for compatibility with cases such as ideals in
        number fields.

        .. TODO::

            Implement this method.

        EXAMPLES::

            sage: R.<t> = GF(9, names='a')[]                                            # needs sage.rings.finite_rings
            sage: I = R.ideal(t^4 + t + 1)                                              # needs sage.rings.finite_rings
            sage: I.absolute_norm()                                                     # needs sage.rings.finite_rings
            Traceback (most recent call last):
            ...
            NotImplementedError
        """
        raise NotImplementedError

    def _macaulay2_init_(self, macaulay2=None):
        """
        Return Macaulay2 ideal corresponding to this ideal.

        EXAMPLES:

        Ideals in multivariate polynomial rings::

            sage: R.<x,y,z,w> = PolynomialRing(ZZ, 4)
            sage: I = R.ideal([x*y-z^2, y^2-w^2]); I
            Ideal (x*y - z^2, y^2 - w^2) of Multivariate Polynomial Ring in x, y, z, w over Integer Ring
            sage: macaulay2(I)                                  # optional - macaulay2
                          2   2    2
            ideal (x*y - z , y  - w )

        Ideals in univariate polynomial rings::

            sage: R.<x> = PolynomialRing(ZZ)
            sage: I = R.ideal([4 + 3*x + x^2, 1 + x^2]); I
            Ideal (x^2 + 3*x + 4, x^2 + 1) of Univariate Polynomial Ring in x over Integer Ring
            sage: macaulay2(I)                                  # optional - macaulay2
                    2            2
            ideal (x  + 3x + 4, x  + 1)

        Field ideals generated from the polynomial ring over
        two variables in the finite field of size 2::

            sage: P.<x,y> = PolynomialRing(GF(2), 2)
            sage: I = sage.rings.ideal.FieldIdeal(P); I
            Ideal (x^2 + x, y^2 + y) of Multivariate Polynomial Ring in x, y
             over Finite Field of size 2
            sage: macaulay2(I)                          # optional - macaulay2          # needs sage.rings.finite_rings
                    2       2
            ideal (x  + x, y  + y)

        Ideals in PIDs::

            sage: macaulay2(ideal(5))                   # optional - macaulay2
            ideal 5
            sage: J = ideal(QQ(5))
            ...
            sage: macaulay2(J)                          # optional - macaulay2
            ideal 1

        TESTS:

        Check that a cached base ring is used (:issue:`28074`)::

            sage: R.<x,y> = QQ[]
            sage: R1 = macaulay2(R)                        # optional - macaulay2
            sage: _ = macaulay2('ZZ[x,y]')                 # optional - macaulay2
            sage: R2 = macaulay2(R.ideal(y^2 - x)).ring()  # optional - macaulay2
            sage: R1._operator('===', R2)                  # optional - macaulay2
            true
        """
        if macaulay2 is None:
            from sage.interfaces.macaulay2 import macaulay2 as m2_default
            macaulay2 = m2_default

        R = self.ring()
        macaulay2.use(R._macaulay2_(macaulay2))
        gens = [repr(x) for x in self.gens()]
        if len(gens) == 0:
            gens = ['0']
        return macaulay2.ideal(gens)

    def free_resolution(self, *args, **kwds):
        r"""
        Return a free resolution of ``self``.

        For input options, see
        :class:`~sage.homology.free_resolution.FreeResolution`.

        EXAMPLES::

            sage: R.<x> = PolynomialRing(QQ)
            sage: I = R.ideal([x^4 + 3*x^2 + 2])
            sage: I.free_resolution()                                                   # needs sage.modules
            S^1 <-- S^1 <-- 0
        """
        if not self.is_principal():
            raise NotImplementedError("the ideal must be a principal ideal")
        from sage.homology.free_resolution import FiniteFreeResolution_free_module
        return FiniteFreeResolution_free_module(self, *args, **kwds)

    def graded_free_resolution(self, *args, **kwds):
        r"""
        Return a graded free resolution of ``self``.

        For input options, see
        :class:`~sage.homology.graded_resolution.GradedFiniteFreeResolution`.

        EXAMPLES::

            sage: R.<x> = PolynomialRing(QQ)
            sage: I = R.ideal([x^3])
            sage: I.graded_free_resolution()                                            # needs sage.modules
            S(0) <-- S(-3) <-- 0
        """
        from sage.homology.graded_resolution import GradedFiniteFreeResolution_free_module
        return GradedFiniteFreeResolution_free_module(self, *args, **kwds)


class Ideal_principal(Ideal_generic):
    """
    A principal ideal.

    See :func:`Ideal()`.
    """
    # now Ideal_principal takes a list.
    #def __init__(self, ring, gen):
    #    Ideal_generic.__init__(self, ring, [gen])

    def _repr_(self):
        """
        Return a string representation of ``self``.

        EXAMPLES::

            sage: R.<x> = ZZ[]
            sage: I = R.ideal(x)
            sage: I # indirect doctest
            Principal ideal (x) of Univariate Polynomial Ring in x over Integer Ring
        """
        return "Principal ideal (%s) of %s" % (self.gen(), self.ring())

    def is_principal(self):
        r"""
        Return ``True`` if the ideal is principal in the ring containing the
        ideal. When the ideal construction is explicitly principal (i.e.
        when we define an ideal with one element) this is always the case.

        EXAMPLES:

        Note that Sage automatically coerces ideals into
        principal ideals during initialization::

            sage: R.<x> = ZZ[]
            sage: I = R.ideal(x)
            sage: J = R.ideal(2,x)
            sage: K = R.base_extend(QQ).ideal(2,x)
            sage: I
            Principal ideal (x) of Univariate Polynomial Ring in x
            over Integer Ring
            sage: J
            Ideal (2, x) of Univariate Polynomial Ring in x over Integer Ring
            sage: K
            Principal ideal (1) of Univariate Polynomial Ring in x
            over Rational Field
            sage: I.is_principal()
            True
            sage: K.is_principal()
            True
        """
        return True

    def gen(self, i=0):
        r"""
        Return the generator of the principal ideal.

        The generator is an element of the ring containing the ideal.

        EXAMPLES:

        A simple example in the integers::

            sage: R = ZZ
            sage: I = R.ideal(7)
            sage: J = R.ideal(7, 14)
            sage: I.gen(); J.gen()
            7
            7

        Note that the generator belongs to the ring from which the ideal
        was initialized::

            sage: R.<x> = ZZ[]
            sage: I = R.ideal(x)
            sage: J = R.base_extend(QQ).ideal(2,x)
            sage: a = I.gen(); a
            x
            sage: b = J.gen(); b
            1
            sage: a.base_ring()
            Integer Ring
            sage: b.base_ring()
            Rational Field
        """
        if i:
            raise ValueError(f"i (={i}) must be 0")
        return self.gens()[0]

    def __contains__(self, x):
        """
        Return ``True`` if ``x`` is in ``self``.

        EXAMPLES::

            sage: P.<x> = PolynomialRing(ZZ)
            sage: I = P.ideal(x^2-2)
            sage: x^2 in I
            False
            sage: x^2-2 in I
            True
            sage: x^2-3 in I
            False
        """
        if self.gen().is_zero():
            return x.is_zero()
        try:
            return self.gen().divides(x)
        except NotImplementedError:
            return self._contains_(self.ring()(x))

    def __hash__(self):
        r"""
        Very stupid constant hash function!

        TESTS::

            sage: P.<x, y> = PolynomialRing(ZZ)
            sage: I = P.ideal(x^2)
            sage: J = [x, y^2 + x*y]*P
            sage: hash(I)
            0
            sage: hash(J)
            0
        """
        return 0

    def _richcmp_(self, other, op):
        """
        Compare the two ideals.

        EXAMPLES:

        Comparison with non-principal ideal::

            sage: R.<x> = ZZ[]
            sage: I = R.ideal([x^3 + 4*x - 1, x + 6])
            sage: J = [x^2] * R
            sage: I > J  # indirect doctest
            True
            sage: J < I  # indirect doctest
            True

        Between two principal ideals::

            sage: P.<x> = PolynomialRing(ZZ)
            sage: I = P.ideal(x^2-2)
            sage: I2 = P.ideal(0)
            sage: I2.is_zero()
            True
            sage: I2 < I
            True
            sage: I3 = P.ideal(x)
            sage: I > I3
            True
        """
        if not isinstance(other, Ideal_generic):
            other = self.ring().ideal(other)

        try:
            if not other.is_principal():
                return rich_to_bool(op, -1)
        except NotImplementedError:
            # If we do not know if the other is principal or not, then we
            #   fallback to the generic implementation
            return Ideal_generic._richcmp_(self, other, op)

        if self.is_zero():
            if not other.is_zero():
                return rich_to_bool(op, -1)
            return rich_to_bool(op, 0)

        # is other.gen() / self.gen() a unit in the base ring?
        g0 = other.gen()
        g1 = self.gen()
        if g0.divides(g1) and g1.divides(g0):
            return rich_to_bool(op, 0)
        return rich_to_bool(op, 1)

    def divides(self, other):
        """
        Return ``True`` if ``self`` divides ``other``.

        EXAMPLES::

            sage: P.<x> = PolynomialRing(QQ)
            sage: I = P.ideal(x)
            sage: J = P.ideal(x^2)
            sage: I.divides(J)
            True
            sage: J.divides(I)
            False
        """
        if isinstance(other, Ideal_principal):
            return self.gen().divides(other.gen())
        raise NotImplementedError


class Ideal_pid(Ideal_principal):
    """
    An ideal of a principal ideal domain.

    See :func:`Ideal()`.

    EXAMPLES::

        sage: I = 8*ZZ
        sage: I
        Principal ideal (8) of Integer Ring
    """
    def __add__(self, other):
        """
        Add the two ideals.

        EXAMPLES::

            sage: I = 8*ZZ
            sage: I2 = 3*ZZ
            sage: I + I2
            Principal ideal (1) of Integer Ring
        """
        if not isinstance(other, Ideal_generic):
            other = self.ring().ideal(other)
        return self.ring().ideal(self.gcd(other))

    def reduce(self, f):
        """
        Return the reduction of `f` modulo ``self``.

        EXAMPLES::

            sage: I = 8*ZZ
            sage: I.reduce(10)
            2
            sage: n = 10; n.mod(I)
            2
        """
        f = self.ring()(f)
        if self.gen() == 0:
            return f
        q, r = f.quo_rem(self.gen())
        return r

    def gcd(self, other):
        r"""
        Return the greatest common divisor of the principal ideal with the
        ideal ``other``; that is, the largest principal ideal
        contained in both the ideal and ``other``

        .. TODO::

            This is not implemented in the case when ``other`` is neither
            principal nor when the generator of ``self`` is contained in
            ``other``. Also, it seems that this class is used only in PIDs--is
            this redundant?

        .. NOTE::

            The second example is broken.

        EXAMPLES:

        An example in the principal ideal domain `\ZZ`::

            sage: R = ZZ
            sage: I = R.ideal(42)
            sage: J = R.ideal(70)
            sage: I.gcd(J)
            Principal ideal (14) of Integer Ring
            sage: J.gcd(I)
            Principal ideal (14) of Integer Ring

        TESTS:

        We cannot take the gcd of a principal ideal with a
        non-principal ideal as well: ( ``gcd(I,J)`` should be `(7)` )

        ::

            sage: R.<x> = ZZ[]
            sage: I = ZZ.ideal(7)
            sage: J = R.ideal(7,x)
            sage: I.gcd(J)
            Traceback (most recent call last):
            ...
            NotImplementedError
            sage: J.gcd(I)
            Traceback (most recent call last):
            ...
            AttributeError: 'Ideal_generic' object has no attribute 'gcd'...

        Note::

            sage: type(I)
            <class 'sage.rings.ideal.Ideal_pid'>
            sage: type(J)
            <class 'sage.rings.ideal.Ideal_generic'>
        """
        if isinstance(other, Ideal_principal):
            return self.ring().ideal(self.gen().gcd(other.gen()))
        elif self.gen() in other:
            return other
        else:
            raise NotImplementedError

    def is_prime(self):
        """
        Return ``True`` if the ideal is prime.

        This relies on the ring elements having a method ``is_irreducible()``
        implemented, since an ideal `(a)` is prime iff `a` is irreducible
        (or 0).

        EXAMPLES::

            sage: ZZ.ideal(2).is_prime()
            True
            sage: ZZ.ideal(-2).is_prime()
            True
            sage: ZZ.ideal(4).is_prime()
            False
            sage: ZZ.ideal(0).is_prime()
            True
            sage: R.<x> = QQ[]
            sage: P = R.ideal(x^2 + 1); P
            Principal ideal (x^2 + 1) of Univariate Polynomial Ring in x over Rational Field
            sage: P.is_prime()                                                          # needs sage.libs.pari
            True

        In fields, only the zero ideal is prime::

            sage: RR.ideal(0).is_prime()
            True
            sage: RR.ideal(7).is_prime()
            False
        """
        if self.is_zero(): # PIDs are integral domains by definition
            return True
        g = self.gen()
        if g.is_one():     # The ideal (1) is never prime
            return False
        if hasattr(g, 'is_irreducible'):
            return g.is_irreducible()

        raise NotImplementedError

    def is_maximal(self):
        """
        Return whether this ideal is maximal.

        Principal ideal domains have Krull dimension 1 (or 0), so an ideal is
        maximal if and only if it's prime (and nonzero if the ring is not a
        field).

        EXAMPLES::

            sage: # needs sage.rings.finite_rings
            sage: R.<t> = GF(5)[]
            sage: p = R.ideal(t^2 + 2)
            sage: p.is_maximal()
            True
            sage: p = R.ideal(t^2 + 1)
            sage: p.is_maximal()
            False
            sage: p = R.ideal(0)
            sage: p.is_maximal()
            False
            sage: p = R.ideal(1)
            sage: p.is_maximal()
            False
        """
        if not self.ring().is_field() and self.is_zero():
            return False
        return self.is_prime()

    def residue_field(self):
        r"""
        Return the residue class field of this ideal, which must be prime.

        .. TODO::

            Implement this for more general rings. Currently only defined
            for `\ZZ` and for number field orders.

        EXAMPLES::

            sage: # needs sage.libs.pari
            sage: P = ZZ.ideal(61); P
            Principal ideal (61) of Integer Ring
            sage: F = P.residue_field(); F
            Residue field of Integers modulo 61
            sage: pi = F.reduction_map(); pi
            Partially defined reduction map:
              From: Rational Field
              To:   Residue field of Integers modulo 61
            sage: pi(123/234)
            6
            sage: pi(1/61)
            Traceback (most recent call last):
            ...
            ZeroDivisionError: Cannot reduce rational 1/61 modulo 61: it has negative valuation
            sage: lift = F.lift_map(); lift
            Lifting map:
              From: Residue field of Integers modulo 61
              To:   Integer Ring
            sage: lift(F(12345/67890))
            33
            sage: (12345/67890) % 61
            33

        TESTS::

            sage: ZZ.ideal(96).residue_field()
            Traceback (most recent call last):
            ...
            ValueError: The ideal (Principal ideal (96) of Integer Ring) is not prime

        ::

            sage: R.<x> = QQ[]
            sage: I = R.ideal(x^2 + 1)
            sage: I.is_prime()                                                          # needs sage.libs.pari
            True
            sage: I.residue_field()
            Traceback (most recent call last):
            ...
            TypeError: residue fields only supported for polynomial rings over finite fields.
        """
        if not self.is_prime():
            raise ValueError("The ideal (%s) is not prime" % self)
        from sage.rings.integer_ring import ZZ
        if self.ring() is ZZ:
            return ZZ.residue_field(self, check=False)
        raise NotImplementedError("residue_field() is only implemented for ZZ and rings of integers of number fields.")

    def radical(self):
        r"""
        Return the radical of this ideal.

        EXAMPLES::

            sage: ZZ.ideal(12).radical()
            Principal ideal (6) of Integer Ring
        """
        return self.ring().ideal(self.gen().radical())


class Ideal_fractional(Ideal_generic):
    """
    Fractional ideal of a ring.

    See :func:`Ideal()`.
    """
    def _repr_(self):
        """
        Return a string representation of ``self``.

        EXAMPLES::

            sage: from sage.rings.ideal import Ideal_fractional
            sage: x = polygen(ZZ, 'x')
            sage: K.<a> = NumberField(x^2 + 1)                                          # needs sage.rings.number_field
            sage: Ideal_fractional(K, [a])  # indirect doctest                          # needs sage.rings.number_field
            Fractional ideal (a) of Number Field in a with defining polynomial x^2 + 1
        """
        return "Fractional ideal %s of %s" % (self._repr_short(), self.ring())

# constructors for standard (benchmark) ideals, written uppercase as
# these are constructors


def Cyclic(R, n=None, homog=False, singular=None):
    """
    Ideal of cyclic ``n``-roots from 1-st ``n`` variables of ``R`` if ``R`` is
    coercible to :class:`Singular <sage.interfaces.singular.Singular>`.

    INPUT:

    - ``R`` -- base ring to construct ideal for

    - ``n`` -- number of cyclic roots (default: ``None``); if ``None``, then
      ``n`` is set to ``R.ngens()``

    - ``homog`` -- boolean (default: ``False``); if ``True`` a homogeneous
      ideal is returned using the last variable in the ideal

    - ``singular`` -- Singular instance to use

    .. NOTE::

       ``R`` will be set as the active ring in
       :class:`Singular <sage.interfaces.singular.Singular>`

    EXAMPLES:

    An example from a multivariate polynomial ring over the
    rationals::

        sage: P.<x,y,z> = PolynomialRing(QQ, 3, order='lex')
        sage: I = sage.rings.ideal.Cyclic(P); I                                         # needs sage.libs.singular
        Ideal (x + y + z, x*y + x*z + y*z, x*y*z - 1)
         of Multivariate Polynomial Ring in x, y, z over Rational Field
        sage: I.groebner_basis()                                                        # needs sage.libs.singular
        [x + y + z, y^2 + y*z + z^2, z^3 - 1]

    We compute a Groebner basis for cyclic 6, which is a standard
    benchmark and test ideal::

        sage: R.<x,y,z,t,u,v> = QQ['x,y,z,t,u,v']
        sage: I = sage.rings.ideal.Cyclic(R, 6)                                         # needs sage.libs.singular
        sage: B = I.groebner_basis()                                                    # needs sage.libs.singular
        sage: len(B)                                                                    # needs sage.libs.singular
        45
    """
    from .rational_field import RationalField

    if n:
        if n > R.ngens():
            raise ArithmeticError("n must be <= R.ngens()")
    else:
        n = R.ngens()

    if singular is None:
        from sage.interfaces.singular import singular as singular_default
        singular = singular_default

    singular.lib("polylib")
    R2 = R.change_ring(RationalField())
    R2._singular_().set_ring()

    if not homog:
        I = singular.cyclic(n)
    else:
        I = singular.cyclic(n).homog(R2.gen(n-1))
    return R2.ideal(I).change_ring(R)


def Katsura(R, n=None, homog=False, singular=None):
    r"""
    `n`-th katsura ideal of `R` if `R` is coercible to
    :class:`Singular <sage.interfaces.singular.Singular>`.

    INPUT:

    - ``R`` -- base ring to construct ideal for

    - ``n`` -- (default: ``None``) which katsura ideal of `R`. If ``None``,
      then ``n`` is set to ``R.ngens()``

    - ``homog`` -- boolean (default: ``False``); if ``True`` a homogeneous
      ideal is returned using the last variable in the ideal

    - ``singular`` -- Singular instance to use

    EXAMPLES::

        sage: P.<x,y,z> = PolynomialRing(QQ, 3)
        sage: I = sage.rings.ideal.Katsura(P, 3); I                                     # needs sage.libs.singular
        Ideal (x + 2*y + 2*z - 1, x^2 + 2*y^2 + 2*z^2 - x, 2*x*y + 2*y*z - y)
        of Multivariate Polynomial Ring in x, y, z over Rational Field

    ::

        sage: Q.<x> = PolynomialRing(QQ, implementation='singular')                     # needs sage.libs.singular
        sage: J = sage.rings.ideal.Katsura(Q,1); J                                      # needs sage.libs.singular
        Ideal (x - 1) of Multivariate Polynomial Ring in x over Rational Field
    """
    from .rational_field import RationalField
    if n:
        if n > R.ngens():
            raise ArithmeticError("n must be <= R.ngens().")
    else:
        n = R.ngens()

    if singular is None:
        from sage.interfaces.singular import singular as singular_default
        singular = singular_default
    singular.lib("polylib")
    R2 = R.change_ring(RationalField())
    R2._singular_().set_ring()

    if not homog:
        I = singular.katsura(n)
    else:
        I = singular.katsura(n).homog(R2.gen(n-1))
    return R2.ideal(I).change_ring(R)


def FieldIdeal(R):
    r"""
    Let ``q = R.base_ring().order()`` and `(x_0,...,x_n)` ``= R.gens()`` then
    if `q` is finite this constructor returns

    .. MATH::

        \langle x_0^q - x_0, ... , x_n^q - x_n \rangle.

    We call this ideal the field ideal and the generators the field
    equations.

    EXAMPLES:

    The field ideal generated from the polynomial ring over
    two variables in the finite field of size 2::

        sage: P.<x,y> = PolynomialRing(GF(2), 2)
        sage: I = sage.rings.ideal.FieldIdeal(P); I
        Ideal (x^2 + x, y^2 + y) of
         Multivariate Polynomial Ring in x, y over Finite Field of size 2

    Another, similar example::

        sage: Q.<x1,x2,x3,x4> = PolynomialRing(GF(2^4, name='alpha'), 4)                # needs sage.rings.finite_rings
        sage: J = sage.rings.ideal.FieldIdeal(Q); J                                     # needs sage.rings.finite_rings
        Ideal (x1^16 + x1, x2^16 + x2, x3^16 + x3, x4^16 + x4) of
         Multivariate Polynomial Ring in x1, x2, x3, x4
          over Finite Field in alpha of size 2^4
    """
    q = R.base_ring().order()
    import sage.rings.infinity
    if q is sage.rings.infinity.infinity:
        raise TypeError("Cannot construct field ideal for R.base_ring().order()==infinity")
    return R.ideal([x**q - x for x in R.gens()])



# sage_setup: distribution = sagemath-categories
r"""
Finite dimensional bialgebras with basis
"""
#*****************************************************************************
#  Copyright (C) 2008 Teresa Gomez-Diaz (CNRS) <Teresa.Gomez-Diaz@univ-mlv.fr>
#                2011 Nicolas M. Thiery <nthiery at users.sf.net>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#                  http://www.gnu.org/licenses/
#******************************************************************************


def FiniteDimensionalBialgebrasWithBasis(base_ring):
    """
    The category of finite dimensional bialgebras with a distinguished basis.

    EXAMPLES::

        sage: C = FiniteDimensionalBialgebrasWithBasis(QQ); C
        Category of finite dimensional bialgebras with basis over Rational Field
        sage: sorted(C.super_categories(), key=str)
        [Category of bialgebras with basis over Rational Field,
         Category of finite dimensional algebras with basis over Rational Field]
        sage: C is Bialgebras(QQ).WithBasis().FiniteDimensional()
        True

    TESTS::

        sage: TestSuite(C).run()
    """
    from sage.categories.bialgebras_with_basis import BialgebrasWithBasis
    return BialgebrasWithBasis(base_ring).FiniteDimensional()



# sage_setup: distribution = sagemath-categories
"""
Signed Tensor Product Functorial Construction

AUTHORS:

- Travis Scrimshaw (2019-07): initial version
"""
# ****************************************************************************
#  Copyright (C) 2019 Travis Scrimshaw <tcscrims at gamil.com>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#                  http://www.gnu.org/licenses/
# ****************************************************************************

from sage.categories.covariant_functorial_construction import CovariantFunctorialConstruction, CovariantConstructionCategory
from sage.typeset.unicode_characters import unicode_otimes


class SignedTensorProductFunctor(CovariantFunctorialConstruction):
    """
    A singleton class for the signed tensor functor.

    This functor takes a collection of graded algebras (possibly with
    basis) and constructs the signed tensor product of those algebras.
    If this algebra is in a subcategory, say that of
    ``Algebras(QQ).Graded()``, it is automatically endowed with
    its natural algebra structure, thanks to the category
    ``Algebras(QQ).Graded().SignedTensorProducts()`` of signed tensor
    products of graded algebras. For elements, it constructs the natural
    tensor product element in the corresponding tensor product of their
    parents.

    The signed tensor functor is covariant: if ``A`` is a subcategory
    of ``B``, then ``A.SignedTensorProducts()`` is a subcategory of
    ``B.SignedTensorProducts()`` (see also
    :class:`~sage.categories.covariant_functorial_construction.CovariantFunctorialConstruction`). Hence,
    the role of ``Algebras(QQ).Graded().SignedTensorProducts()`` is solely
    to provide mathematical information and algorithms which are relevant to
    signed tensor product of graded algebras.

    Those are implemented in the nested class
    :class:`~sage.categories.algebras.Algebras.SignedTensorProducts`
    of ``Algebras(QQ).Graded()``. This nested class is itself a subclass of
    :class:`~sage.categories.signed_tensor.SignedTensorProductsCategory`.

    EXAMPLES::

        sage: tensor_signed
        The signed tensor functorial construction

    TESTS::

        sage: TestSuite(tensor_signed).run()
    """
    _functor_name = "tensor"
    _functor_category = "SignedTensorProducts"
    symbol = " # "
    unicode_symbol = f" {unicode_otimes} "

    def _repr_(self):
        """
        EXAMPLES::

            sage: tensor_signed
            The signed tensor functorial construction
        """
        # We need this to distinguish it from tensor(), but we want
        #   it to have the same _functor_name (which is used in the
        #   default _repr_
        return "The signed tensor functorial construction"


tensor_signed = SignedTensorProductFunctor()


class SignedTensorProductsCategory(CovariantConstructionCategory):
    r"""
    An abstract base class for all SignedTensorProducts's categories.

    TESTS::

        sage: C = AlgebrasWithBasis(QQ).Graded().SignedTensorProducts()
        sage: C
        Category of signed tensor products of graded algebras with basis over Rational Field
        sage: C.base_category()
        Category of graded algebras with basis over Rational Field
        sage: latex(C)
        \mathbf{SignedTensorProducts}(\mathbf{GradedAlgebrasWithBasis}(\mathbf{AlgebrasWithBasis}_{\Bold{Q}}))
        sage: TestSuite(C).run()
    """
    _functor_category = "SignedTensorProducts"

    def SignedTensorProducts(self):
        r"""
        Return the category of signed tensor products of objects of ``self``.

        By associativity of signed tensor products, this is ``self`` (a tensor
        product of signed tensor products of `Cat`'s is a tensor product of
        `Cat`'s with the same twisting morphism)

        EXAMPLES::

            sage: AlgebrasWithBasis(QQ).Graded().SignedTensorProducts().SignedTensorProducts()
            Category of signed tensor products of graded algebras with basis
             over Rational Field
        """
        return self

    def base(self):
        """
        The base of a signed tensor product is the base (usually a ring)
        of the underlying category.

        EXAMPLES::

            sage: AlgebrasWithBasis(ZZ).Graded().SignedTensorProducts().base()
            Integer Ring
        """
        return self.base_category().base()



# sage_setup: distribution = sagemath-categories
r"""
Left modules
"""
#*****************************************************************************
#  Copyright (C) 2008 Teresa Gomez-Diaz (CNRS) <Teresa.Gomez-Diaz@univ-mlv.fr>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#                  http://www.gnu.org/licenses/
#******************************************************************************

from sage.categories.category_types import Category_over_base_ring
from sage.categories.commutative_additive_groups import CommutativeAdditiveGroups


#?class LeftModules(Category_over_base_rng):
class LeftModules(Category_over_base_ring):
    """
    The category of left modules
    left modules over an rng (ring not necessarily with unit), i.e.
    an abelian group with left multiplication by elements of the rng

    EXAMPLES::

        sage: LeftModules(ZZ)
        Category of left modules over Integer Ring
        sage: LeftModules(ZZ).super_categories()
        [Category of commutative additive groups]

    TESTS::

        sage: TestSuite(LeftModules(ZZ)).run()
    """

    def super_categories(self):
        """
        EXAMPLES::

            sage: LeftModules(QQ).super_categories()
            [Category of commutative additive groups]
        """
        return [CommutativeAdditiveGroups()]

    class ParentMethods:
        pass

    class ElementMethods:
        ## r * x
        pass



# sage_setup: distribution = sagemath-categories
r"""
Bialgebras with basis
"""
# ****************************************************************************
#  Copyright (C) 2008 Teresa Gomez-Diaz (CNRS) <Teresa.Gomez-Diaz@univ-mlv.fr>
#  Copyright (C) 2008-2011 Nicolas M. Thiery <nthiery at users.sf.net>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#                  https://www.gnu.org/licenses/
# *****************************************************************************

from sage.categories.category_with_axiom import CategoryWithAxiom_over_base_ring
from sage.categories.tensor import tensor
from sage.misc.superseded import deprecated_function_alias


class BialgebrasWithBasis(CategoryWithAxiom_over_base_ring):
    r"""
    The category of bialgebras with a distinguished basis.

    EXAMPLES::

        sage: C = BialgebrasWithBasis(QQ); C
        Category of bialgebras with basis over Rational Field

        sage: sorted(C.super_categories(), key=str)
        [Category of algebras with basis over Rational Field,
         Category of bialgebras over Rational Field,
         Category of coalgebras with basis over Rational Field]

    TESTS::

        sage: TestSuite(BialgebrasWithBasis(ZZ)).run()
    """
    class ParentMethods:

        def convolution_product(self, *maps):
            r"""
            Return the convolution product (a map) of the given maps.

            Let `A` and `B` be bialgebras over a commutative ring `R`.
            Given maps `f_i : A \to B` for `1 \leq i < n`, define the
            convolution product

            .. MATH::

                (f_1 * f_2 * \cdots * f_n) := \mu^{(n-1)} \circ (f_1 \otimes
                f_2 \otimes \cdots \otimes f_n) \circ \Delta^{(n-1)},

            where `\Delta^{(k)} := \bigl(\Delta \otimes
            \mathrm{Id}^{\otimes(k-1)}\bigr) \circ \Delta^{(k-1)}`,
            with `\Delta^{(1)} = \Delta` (the ordinary coproduct in `A`) and
            `\Delta^{(0)} = \mathrm{Id}`; and with `\mu^{(k)} := \mu \circ
            \bigl(\mu^{(k-1)} \otimes \mathrm{Id})` and `\mu^{(1)} = \mu`
            (the ordinary product in `B`). See [Swe1969]_.

            (In the literature, one finds, e.g., `\Delta^{(2)}` for what we
            denote above as `\Delta^{(1)}`. See [KMN2012]_.)

            INPUT:

            - ``maps`` -- any number `n \geq 0` of linear maps `f_1, f_2,
              \ldots, f_n` on ``self``; or a single ``list`` or ``tuple``
              of such maps

            OUTPUT:

            - the new map `f_1 * f_2 * \cdots * f_2` representing their
              convolution product

            .. SEEALSO::

                :meth:`sage.categories.bialgebras.ElementMethods.convolution_product`

            AUTHORS:

            - Aaron Lauve - 12 June 2015 - Sage Days 65

            .. TODO::

                Remove dependency on ``modules_with_basis`` methods.

            EXAMPLES:

            We construct some maps: the identity, the antipode and
            projection onto the homogeneous component of degree 2::

                sage: Id = lambda x: x
                sage: Antipode = lambda x: x.antipode()
                sage: Proj2 = lambda x: x.parent().sum_of_terms([(m, c) for (m, c) in x if m.size() == 2])

            Compute the convolution product of the identity with itself and
            with the projection ``Proj2`` on the Hopf algebra of
            non-commutative symmetric functions::

                sage: # needs sage.combinat sage.modules
                sage: R = NonCommutativeSymmetricFunctions(QQ).ribbon()
                sage: T = R.convolution_product([Id, Id])
                sage: [T(R(comp)) for comp in Compositions(3)]
                [4*R[1, 1, 1] + R[1, 2] + R[2, 1],
                 2*R[1, 1, 1] + 4*R[1, 2] + 2*R[2, 1] + 2*R[3],
                 2*R[1, 1, 1] + 2*R[1, 2] + 4*R[2, 1] + 2*R[3],
                 R[1, 2] + R[2, 1] + 4*R[3]]
                sage: T = R.convolution_product(Proj2, Id)
                sage: [T(R([i])) for i in range(1, 5)]
                [0, R[2], R[2, 1] + R[3], R[2, 2] + R[4]]

            Compute the convolution product of no maps on the Hopf algebra of
            symmetric functions in non-commuting variables. This is the
            composition of the counit with the unit::

                sage: # needs sage.combinat sage.graphs sage.modules
                sage: m = SymmetricFunctionsNonCommutingVariables(QQ).m()
                sage: T = m.convolution_product()
                sage: [T(m(lam))
                ....:  for lam in SetPartitions(0).list() + SetPartitions(2).list()]
                [m{}, 0, 0]

            Compute the convolution product of the projection ``Proj2`` with
            the identity on the Hopf algebra of symmetric functions in
            non-commuting variables::

                sage: T = m.convolution_product(Proj2, Id)                              # needs sage.combinat sage.graphs sage.modules
                sage: [T(m(lam)) for lam in SetPartitions(3)]                           # needs sage.combinat sage.graphs sage.modules
                [0,
                 m{{1, 2}, {3}} + m{{1, 2, 3}},
                 m{{1, 2}, {3}} + m{{1, 2, 3}},
                 m{{1, 2}, {3}} + m{{1, 2, 3}},
                 3*m{{1}, {2}, {3}} + 3*m{{1}, {2, 3}} + 3*m{{1, 3}, {2}}]

            Compute the convolution product of the antipode with itself and the
            identity map on group algebra of the symmetric group::

                sage: # needs sage.combinat sage.groups sage.modules
                sage: G = SymmetricGroup(3)
                sage: QG = GroupAlgebra(G, QQ)
                sage: x = QG.sum_of_terms(
                ....:         [(p, p.number_of_peaks() + p.number_of_inversions())
                ....:          for p in Permutations(3)]
                ....:     ); x
                2*[1, 3, 2] + [2, 1, 3] + 3*[2, 3, 1] + 2*[3, 1, 2] + 3*[3, 2, 1]
                sage: T = QG.convolution_product(Antipode, Antipode, Id)
                sage: T(x)
                2*[1, 3, 2] + [2, 1, 3] + 2*[2, 3, 1] + 3*[3, 1, 2] + 3*[3, 2, 1]
            """
            onbasis = lambda x: self.term(x).convolution_product(*maps)
            return self.module_morphism(on_basis=onbasis, codomain=self)

    class ElementMethods:

        def convolution_power_of_id(self, n):
            r"""
            Compute the `n`-th convolution power of the identity morphism
            `\mathrm{Id}` on ``self``.

            INPUT:

            - ``n`` -- nonnegative integer

            OUTPUT: the image of ``self`` under the convolution power `\mathrm{Id}^{*n}`

            .. NOTE::

                In the literature, this is also called a Hopf power or
                Sweedler power, cf. [AL2015]_.

            .. SEEALSO::

                :meth:`sage.categories.bialgebras.ElementMethods.convolution_product`

            .. TODO::

                Remove dependency on ``modules_with_basis`` methods.

            EXAMPLES::

                sage: # needs sage.combinat sage.modules
                sage: h = SymmetricFunctions(QQ).h()
                sage: h[5].convolution_power_of_id(2)
                2*h[3, 2] + 2*h[4, 1] + 2*h[5]
                sage: h[5].plethysm(2*h[1])
                2*h[3, 2] + 2*h[4, 1] + 2*h[5]
                sage: h([]).convolution_power_of_id(0)
                h[]
                sage: h([]).convolution_power_of_id(1)
                h[]
                sage: h[3,2].convolution_power_of_id(0)
                0
                sage: h[3,2].convolution_power_of_id(1)
                h[3, 2]

            ::

                sage: S = NonCommutativeSymmetricFunctions(QQ).S()                      # needs sage.combinat sage.modules
                sage: S[4].convolution_power_of_id(5)                                   # needs sage.combinat sage.modules
                5*S[1, 1, 1, 1] + 10*S[1, 1, 2] + 10*S[1, 2, 1]
                 + 10*S[1, 3] + 10*S[2, 1, 1] + 10*S[2, 2] + 10*S[3, 1] + 5*S[4]


            ::

                sage: m = SymmetricFunctionsNonCommutingVariables(QQ).m()               # needs sage.combinat sage.graphs sage.modules
                sage: m[[1,3],[2]].convolution_power_of_id(-2)                          # needs sage.combinat sage.graphs sage.modules
                3*m{{1}, {2, 3}} + 3*m{{1, 2}, {3}} + 6*m{{1, 2, 3}} - 2*m{{1, 3}, {2}}
            """
            if n < 0:
                if hasattr(self, 'antipode'):
                    T = lambda x: x.antipode()
                    n = abs(n)
                else:
                    raise ValueError("antipode not defined; cannot take negative convolution powers: {} < 0".format(n))
            else:
                T = lambda x: x
            return self.convolution_product([T] * n)

        adams_operator = deprecated_function_alias(36396,
                                                   convolution_power_of_id)

        def convolution_product(self, *maps):
            r"""
            Return the image of ``self`` under the convolution product (map) of
            the maps.

            Let `A` and `B` be bialgebras over a commutative ring `R`.
            Given maps `f_i : A \to B` for `1 \leq i < n`, define the
            convolution product

            .. MATH::

                (f_1 * f_2 * \cdots * f_n) := \mu^{(n-1)} \circ (f_1 \otimes
                f_2 \otimes \cdots \otimes f_n) \circ \Delta^{(n-1)},

            where `\Delta^{(k)} := \bigl(\Delta \otimes
            \mathrm{Id}^{\otimes(k-1)}\bigr) \circ \Delta^{(k-1)}`,
            with `\Delta^{(1)} = \Delta` (the ordinary coproduct in `A`) and
            `\Delta^{(0)} = \mathrm{Id}`; and with `\mu^{(k)} := \mu \circ
            \bigl(\mu^{(k-1)} \otimes \mathrm{Id})` and `\mu^{(1)} = \mu`
            (the ordinary product in `B`). See [Swe1969]_.

            (In the literature, one finds, e.g., `\Delta^{(2)}` for what we
            denote above as `\Delta^{(1)}`. See [KMN2012]_.)

            INPUT:

            - ``maps`` -- any number `n \geq 0` of linear maps `f_1, f_2,
              \ldots, f_n` on ``self.parent()``; or a single ``list`` or
              ``tuple`` of such maps

            OUTPUT: the convolution product of ``maps`` applied to ``self``

            AUTHORS:

            - Amy Pang - 12 June 2015 - Sage Days 65

            .. TODO::

                Remove dependency on ``modules_with_basis`` methods.

            EXAMPLES:

            We compute convolution products of the identity and antipode maps
            on Schur functions::

                sage: # needs sage.combinat sage.modules
                sage: Id = lambda x: x
                sage: Antipode = lambda x: x.antipode()
                sage: s = SymmetricFunctions(QQ).schur()
                sage: s[3].convolution_product(Id, Id)                                  # needs lrcalc_python
                2*s[2, 1] + 4*s[3]
                sage: s[3,2].convolution_product(Id) == s[3,2]
                True

            The method accepts multiple arguments, or a single argument
            consisting of a list of maps::

                sage: s[3,2].convolution_product(Id, Id)                                # needs lrcalc_python sage.combinat sage.modules
                2*s[2, 1, 1, 1] + 6*s[2, 2, 1] + 6*s[3, 1, 1] + 12*s[3, 2] + 6*s[4, 1] + 2*s[5]
                sage: s[3,2].convolution_product([Id, Id])                              # needs lrcalc_python sage.combinat sage.modules
                2*s[2, 1, 1, 1] + 6*s[2, 2, 1] + 6*s[3, 1, 1] + 12*s[3, 2] + 6*s[4, 1] + 2*s[5]

            We test the defining property of the antipode morphism; namely,
            that the antipode is the inverse of the identity map in the
            convolution algebra whose identity element is the composition of
            the counit and unit::

                sage: (s[3,2].convolution_product()                                     # needs lrcalc_python sage.combinat sage.modules
                ....:   == s[3,2].convolution_product(Antipode, Id)
                ....:   == s[3,2].convolution_product(Id, Antipode))
                True

            ::

                sage: Psi = NonCommutativeSymmetricFunctions(QQ).Psi()                  # needs sage.combinat sage.modules
                sage: Psi[2,1].convolution_product(Id, Id, Id)                          # needs sage.combinat sage.modules
                3*Psi[1, 2] + 6*Psi[2, 1]
                sage: (Psi[5,1] - Psi[1,5]).convolution_product(Id, Id, Id)             # needs sage.combinat sage.modules
                -3*Psi[1, 5] + 3*Psi[5, 1]

            ::

                sage: # needs sage.combinat sage.groups sage.modules
                sage: G = SymmetricGroup(3)
                sage: QG = GroupAlgebra(G, QQ)
                sage: x = QG.sum_of_terms([(p, p.length())
                ....:                      for p in Permutations(3)]); x
                [1, 3, 2] + [2, 1, 3] + 2*[2, 3, 1] + 2*[3, 1, 2] + 3*[3, 2, 1]
                sage: x.convolution_product(Id, Id)
                5*[1, 2, 3] + 2*[2, 3, 1] + 2*[3, 1, 2]
                sage: x.convolution_product(Id, Id, Id)
                4*[1, 2, 3] + [1, 3, 2] + [2, 1, 3] + 3*[3, 2, 1]
                sage: x.convolution_product([Id] * 6)
                9*[1, 2, 3]

            TESTS::

                sage: Id = lambda x: x
                sage: Antipode = lambda x: x.antipode()

            ::

                sage: # needs sage.combinat sage.modules
                sage: h = SymmetricFunctions(QQ).h()
                sage: h[5].convolution_product([Id, Id])
                2*h[3, 2] + 2*h[4, 1] + 2*h[5]
                sage: h.one().convolution_product([Id, Antipode])
                h[]
                sage: h[3,2].convolution_product([Id, Antipode])
                0
                sage: (h.one().convolution_product([Id, Antipode])
                ....:   == h.one().convolution_product())
                True

            ::

                sage: S = NonCommutativeSymmetricFunctions(QQ).S()                      # needs sage.combinat sage.modules
                sage: S[4].convolution_product([Id] * 5)                                # needs sage.combinat sage.modules
                5*S[1, 1, 1, 1] + 10*S[1, 1, 2] + 10*S[1, 2, 1] + 10*S[1, 3]
                 + 10*S[2, 1, 1] + 10*S[2, 2] + 10*S[3, 1] + 5*S[4]

            ::

                sage: # needs sage.combinat sage.graphs sage.modules
                sage: m = SymmetricFunctionsNonCommutingVariables(QQ).m()
                sage: m[[1,3],[2]].convolution_product([Antipode, Antipode])
                3*m{{1}, {2, 3}} + 3*m{{1, 2}, {3}} + 6*m{{1, 2, 3}} - 2*m{{1, 3}, {2}}
                sage: m[[]].convolution_product([])
                m{}
                sage: m[[1,3],[2]].convolution_product([])
                0

            ::

                sage: # needs sage.combinat sage.groups sage.modules
                sage: QS = SymmetricGroupAlgebra(QQ, 5)
                sage: x = QS.sum_of_terms(zip(Permutations(5)[3:6], [1,2,3])); x
                [1, 2, 4, 5, 3] + 2*[1, 2, 5, 3, 4] + 3*[1, 2, 5, 4, 3]
                sage: x.convolution_product([Antipode, Id])
                6*[1, 2, 3, 4, 5]
                sage: x.convolution_product(Id, Antipode, Antipode, Antipode)
                3*[1, 2, 3, 4, 5] + [1, 2, 4, 5, 3] + 2*[1, 2, 5, 3, 4]

            ::

                sage: # needs sage.combinat sage.groups sage.modules
                sage: G = SymmetricGroup(3)
                sage: QG = GroupAlgebra(G, QQ)
                sage: x = QG.sum_of_terms([(p, p.length())
                ....:                      for p in Permutations(3)]); x
                [1, 3, 2] + [2, 1, 3] + 2*[2, 3, 1] + 2*[3, 1, 2] + 3*[3, 2, 1]
                sage: x.convolution_product(Antipode, Id)
                9*[1, 2, 3]
                sage: x.convolution_product([Id, Antipode, Antipode, Antipode])
                5*[1, 2, 3] + 2*[2, 3, 1] + 2*[3, 1, 2]

            ::

                sage: (s[3,2].counit().parent()                                         # needs sage.combinat sage.modules
                ....:   == s[3,2].convolution_product().parent())
                False
            """
            # Be flexible on how the maps are entered: accept a list/tuple of
            # maps as well as multiple arguments
            if len(maps) == 1 and isinstance(maps[0], (list, tuple)):
                T = tuple(maps[0])
            else:
                T = maps

            H = self.parent()

            n = len(T)
            if n == 0:
                return H.one() * self.counit()
            if n == 1:
                return T[0](self)

            # We apply the maps T_i and products concurrently with coproducts, as this
            # seems to be faster than applying a composition of maps, e.g., (H.nfold_product) * tensor(T) * (H.nfold_coproduct).

            out = tensor((H.one(), self))
            HH = tensor((H, H))

            for mor in T[:-1]:
                # ALGORITHM:
                # ``split_convolve`` moves terms of the form x # y to x*Ti(y1) # y2 in Sweedler notation.
                def split_convolve(x_y):
                    x, y = x_y
                    return (((xy1, y2), c * d)
                            for ((y1, y2), d) in H.term(y).coproduct()
                            for (xy1, c) in H.term(x) * mor(H.term(y1)))
                out = HH.module_morphism(on_basis=lambda t: HH.sum_of_terms(split_convolve(t)), codomain=HH)(out)

            # Apply final map `T_n` to last term, `y`, and multiply.
            return HH.module_morphism(on_basis=lambda xy: H.term(xy[0]) * T[-1](H.term(xy[1])), codomain=H)(out)



# sage_setup: distribution = sagemath-categories
r"""
Poor Man's map
"""
# ****************************************************************************
#       Copyright (C) 2009 Nicolas M. Thiery <nthiery at users.sf.net>
#                     2016 Julian Rüth <julian.rueth@fsfe.org>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#                  https://www.gnu.org/licenses/
# ****************************************************************************
import sage.structure.sage_object


class PoorManMap(sage.structure.sage_object.SageObject):
    """
    A class for maps between sets which are not (yet) modeled by parents.

    Could possibly disappear when all combinatorial classes / enumerated sets will be parents.

    INPUT:

    - ``function`` -- a callable or an iterable of callables. This represents
      the underlying function used to implement this map. If it is an iterable,
      then the callables will be composed to implement this map.

    - ``domain`` -- the domain of this map or ``None`` if the domain is not
      known or should remain unspecified

    - ``codomain`` -- the codomain of this map or ``None`` if the codomain is
      not known or should remain unspecified

    - ``name`` -- a name for this map or ``None`` if this map has no particular
      name

    EXAMPLES::

        sage: from sage.categories.poor_man_map import PoorManMap
        sage: f = PoorManMap(factorial, domain=(1, 2, 3), codomain=(1, 2, 6))
        sage: f
        A map from (1, 2, 3) to (1, 2, 6)
        sage: f(3)
        6

    The composition of several functions can be created by passing in a tuple
    of functions::

        sage: i = PoorManMap((factorial, sqrt), domain=(1, 4, 9), codomain=(1, 2, 6))

    However, the same effect can also be achieved by just composing maps::

        sage: g = PoorManMap(factorial, domain=(1, 2, 3), codomain=(1, 2, 6))
        sage: h = PoorManMap(sqrt, domain=(1, 4, 9), codomain=(1, 2, 3))
        sage: i == g*h
        True
    """
    def __init__(self, function, domain=None, codomain=None, name=None):
        """
        TESTS::

            sage: from sage.categories.poor_man_map import PoorManMap
            sage: f = PoorManMap(factorial, domain=(1, 2, 3), codomain=(1, 2, 6))
            sage: g = PoorManMap(sqrt, domain=(1, 4, 9), codomain=(1, 2, 6))

            sage: TestSuite(f).run()
            sage: TestSuite(f*g).run()
        """
        from collections.abc import Iterable
        if not isinstance(function, Iterable):
            function = (function,)
        self._functions = tuple(function)
        self._domain = domain
        self._codomain = codomain
        self._name = name

    def _repr_(self):
        """
        EXAMPLES::

            sage: from sage.categories.poor_man_map import PoorManMap
            sage: PoorManMap(lambda x: x+2)  # indirect doctest
            A map
            sage: PoorManMap(lambda x: x+2, domain=(1,2,3))
            A map from (1, 2, 3)
            sage: PoorManMap(lambda x: x+2, domain=(1,2,3))
            A map from (1, 2, 3)
            sage: PoorManMap(lambda x: x+2, codomain=(3,4,5))
            A map to (3, 4, 5)
        """
        return ((self._name if self._name is not None else "A map") +
                (" from %s" % (self._domain,) if self._domain is not None else "") +
                (" to %s" % (self._codomain,) if self._codomain is not None else ""))

    def domain(self):
        """
        Return the domain of ``self``.

        EXAMPLES::

            sage: from sage.categories.poor_man_map import PoorManMap
            sage: PoorManMap(lambda x: x+1, domain=(1,2,3), codomain=(2,3,4)).domain()
            (1, 2, 3)
        """
        return self._domain

    def codomain(self):
        """
        Return the codomain of ``self``.

        EXAMPLES::

            sage: from sage.categories.poor_man_map import PoorManMap
            sage: PoorManMap(lambda x: x+1, domain=(1,2,3), codomain=(2,3,4)).codomain()
            (2, 3, 4)
        """
        return self._codomain

    def __eq__(self, other):
        r"""
        Return whether this map is equal to ``other``.

        EXAMPLES::

            sage: from sage.categories.poor_man_map import PoorManMap
            sage: f = PoorManMap(factorial, domain=(1,2,3), codomain=(1,2,6))
            sage: g = PoorManMap(factorial, domain=(1,2,3), codomain=(1,2,6))
            sage: h1 = PoorManMap(factorial, domain=(1,2,3), codomain=(1,2,6,8))
            sage: h2 = PoorManMap(factorial, domain=(1,2,3), codomain=(1,2,6,8))
            sage: h3 = PoorManMap(factorial, domain=(1,2,3,4), codomain=(1,2,6))
            sage: h4 = PoorManMap(lambda x: x, domain=(1,2,3), codomain=(1,2,6))
            sage: f == g, f == h1, f == h2, f == h3, f == h4, f == 1, 1 == f
            (True, False, False, False, False, False, False)
        """
        if isinstance(other, PoorManMap):
            return (self._functions == other._functions
                    and self._domain == other._domain
                    and self._codomain == other._codomain
                    and self._name == other._name)
        else:
            return False

    def __ne__(self, other):
        r"""
        Return whether this map is not equal to ``other``.

        EXAMPLES::

            sage: from sage.categories.poor_man_map import PoorManMap
            sage: f = PoorManMap(factorial, domain=(1,2,3), codomain=(1,2,6))
            sage: g = PoorManMap(factorial, domain=(1,2,3), codomain=(1,2,6))
            sage: h1 = PoorManMap(factorial, domain=(1,2,3), codomain=(1,2,6,8))
            sage: h2 = PoorManMap(factorial, domain=(1,2,3), codomain=(1,2,6,8))
            sage: h3 = PoorManMap(factorial, domain=(1,2,3,4), codomain=(1,2,6))
            sage: h4 = PoorManMap(lambda x: x, domain=(1,2,3), codomain=(1,2,6))
            sage: f != g, f != h1, f != h2, f != h3, f != h4, f != 1, 1 != f
            (False, True, True, True, True, True, True)
        """
        return not (self == other)

    def __hash__(self):
        r"""
        Return a hash value for this map.

        TESTS::

            sage: from sage.categories.poor_man_map import PoorManMap
            sage: f = PoorManMap(factorial, domain=(1,2,3), codomain=(1,2,6))
            sage: g = PoorManMap(factorial, domain=(1,2,3), codomain=(1,2,6))
            sage: hash(f) == hash(g)
            True
        """
        return hash((self._functions, self._domain, self._codomain, self._name))

    def __mul__(self, other):
        r"""
        Composition.

        INPUT:

        - ``self`` -- a map `f`
        - ``other`` -- a map `g`

        Returns the composition map `f\circ g` of `f` and `g`

        EXAMPLES::

            sage: from sage.categories.poor_man_map import PoorManMap
            sage: f = PoorManMap(lambda x: x+1, domain=(1,2,3), codomain=(2,3,4))
            sage: g = PoorManMap(lambda x: -x,  domain=(2,3,4), codomain=(-2,-3,-4))
            sage: g*f
            A map from (1, 2, 3) to (-2, -3, -4)

        Note that the compatibility of the domains and codomains is for performance
        reasons only checked for proper parents. For example, the incompatibility
        is not detected here::

            sage: f*g
            A map from (2, 3, 4) to (2, 3, 4)

        But it is detected here::

            sage: g = PoorManMap(factorial, domain=ZZ, codomain=ZZ)
            sage: h = PoorManMap(sqrt, domain=RR, codomain=CC)                          # needs sage.rings.real_mpfr
            sage: g*h                                                                   # needs sage.rings.real_mpfr
            Traceback (most recent call last):
            ...
            ValueError: the codomain Complex Field with 53 bits of precision
            does not coerce into the domain Integer Ring
            sage: h*g                                                                   # needs sage.rings.real_mpfr
            A map from Integer Ring to Complex Field with 53 bits of precision
        """
        self_domain = self.domain()

        try:
            other_codomain = other.codomain()
        except AttributeError:
            other_codomain = None

        if self_domain is not None and other_codomain is not None:
            from sage.structure.parent import Parent
            if isinstance(self_domain, Parent) and isinstance(other_codomain, Parent):
                if not self_domain.has_coerce_map_from(other_codomain):
                    raise ValueError("the codomain %r does not coerce into the domain %r" % (other_codomain, self_domain))

        codomain = self.codomain()
        try:
            domain = other.domain()
        except AttributeError:
            domain = None

        if isinstance(other, PoorManMap):
            other = other._functions
        else:
            other = (other,)

        return PoorManMap(self._functions + other, domain=domain, codomain=codomain)

    def __call__(self, *args):
        """
        EXAMPLES::

            sage: from sage.categories.poor_man_map import PoorManMap
            sage: f = PoorManMap(lambda x: x+1, domain=(1,2,3), codomain=(2,3,4))
            sage: f(2)
            3

            sage: g = PoorManMap(lambda x: -x,  domain=(2,3,4), codomain=(-2,-3,-4))
            sage: (g*f)(2)
            -3
        """
        for function in reversed(self._functions):
            args = [function(*args)]
        return args[0]

    def _sympy_(self):
        """
        EXAMPLES::

            sage: from sage.categories.poor_man_map import PoorManMap
            sage: h = PoorManMap(sin, domain=RR, codomain=RR)
            sage: h._sympy_()                                                           # needs sympy sage.symbolic
            sin
        """
        from sympy import Lambda, sympify
        if len(self._functions) == 1:
            return sympify(self._functions[0])
        raise NotImplementedError



# sage_setup: distribution = sagemath-categories
r"""
Complex reflection groups
"""
#*****************************************************************************
#       Copyright (C) 2011-2015 Christian Stump <christian.stump at gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#                  http://www.gnu.org/licenses/
#*****************************************************************************

from sage.misc.cachefunc import cached_method
from sage.misc.lazy_import import LazyImport
from sage.categories.category_singleton import Category_singleton
from sage.categories.complex_reflection_or_generalized_coxeter_groups import ComplexReflectionOrGeneralizedCoxeterGroups


class ComplexReflectionGroups(Category_singleton):
    r"""
    The category of complex reflection groups.

    Let `V` be a complex vector space. A *complex reflection* is an
    element of `\operatorname{GL}(V)` fixing a hyperplane pointwise
    and acting by multiplication by a root of unity on a complementary
    line.

    A *complex reflection group* is a group `W` that is (isomorphic
    to) a subgroup of some general linear group `\operatorname{GL}(V)`
    generated by a distinguished set of complex reflections.

    The dimension of `V` is the *rank* of `W`.

    For a comprehensive treatment of complex reflection groups and
    many definitions and theorems used here, we refer to [LT2009]_.
    See also :wikipedia:`Reflection_group`.

    .. SEEALSO::

        :func:`ReflectionGroup` for usage examples of this category.

    EXAMPLES::

        sage: from sage.categories.complex_reflection_groups import ComplexReflectionGroups
        sage: ComplexReflectionGroups()
        Category of complex reflection groups
        sage: ComplexReflectionGroups().super_categories()
        [Category of complex reflection or generalized Coxeter groups]
        sage: ComplexReflectionGroups().all_super_categories()
        [Category of complex reflection groups,
         Category of complex reflection or generalized Coxeter groups,
         Category of groups,
         Category of monoids,
         Category of finitely generated semigroups,
         Category of semigroups,
         Category of finitely generated magmas,
         Category of inverse unital magmas,
         Category of unital magmas,
         Category of magmas,
         Category of enumerated sets,
         Category of sets,
         Category of sets with partial maps,
         Category of objects]

    An example of a reflection group::

        sage: W = ComplexReflectionGroups().example(); W                                # needs sage.combinat
        5-colored permutations of size 3

    ``W`` is in the category of complex reflection groups::

        sage: W in ComplexReflectionGroups()                                            # needs sage.combinat
        True

    TESTS::

        sage: TestSuite(W).run()                                                        # needs sage.combinat
        sage: TestSuite(ComplexReflectionGroups()).run()
    """

    @cached_method
    def super_categories(self):
        r"""
        Return the super categories of ``self``.

        EXAMPLES::

            sage: from sage.categories.complex_reflection_groups import ComplexReflectionGroups
            sage: ComplexReflectionGroups().super_categories()
            [Category of complex reflection or generalized Coxeter groups]
        """
        return [ComplexReflectionOrGeneralizedCoxeterGroups()]

    def additional_structure(self):
        r"""
        Return ``None``.

        Indeed, all the structure complex reflection groups have in
        addition to groups (simple reflections, ...) is already
        defined in the super category.

        .. SEEALSO:: :meth:`Category.additional_structure`

        EXAMPLES::

            sage: from sage.categories.complex_reflection_groups import ComplexReflectionGroups
            sage: ComplexReflectionGroups().additional_structure()
        """
        return None

    def example(self):
        r"""
        Return an example of a complex reflection group.

        EXAMPLES::

            sage: from sage.categories.complex_reflection_groups import ComplexReflectionGroups
            sage: ComplexReflectionGroups().example()                                   # needs sage.combinat
            5-colored permutations of size 3
        """
        from sage.combinat.colored_permutations import ColoredPermutations
        return ColoredPermutations(5, 3)

    class ParentMethods:

        @cached_method
        def rank(self):
            r"""
            Return the rank of ``self``.

            The rank of ``self`` is the dimension of the smallest
            faithfull reflection representation of ``self``.

            EXAMPLES::

                sage: W = CoxeterGroups().example(); W
                The symmetric group on {0, ..., 3}
                sage: W.rank()
                3
            """

    Finite = LazyImport('sage.categories.finite_complex_reflection_groups',
                        'FiniteComplexReflectionGroups', as_name='Finite')



# sage_setup: distribution = sagemath-categories
"""
Examples of infinite enumerated sets
"""
# ****************************************************************************
#  Copyright (C) 2009 Florent Hivert <Florent.Hivert@univ-rouen.fr>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#                  https://www.gnu.org/licenses/
# ****************************************************************************

from sage.structure.parent import Parent
from sage.categories.infinite_enumerated_sets import InfiniteEnumeratedSets
from sage.structure.unique_representation import UniqueRepresentation
from sage.rings.integer import Integer


class NonNegativeIntegers(UniqueRepresentation, Parent):
    r"""
    An example of infinite enumerated set: the nonnegative integers.

    This class provides a minimal implementation of an infinite enumerated set.

    EXAMPLES::

        sage: NN = InfiniteEnumeratedSets().example()
        sage: NN
        An example of an infinite enumerated set: the nonnegative integers
        sage: NN.cardinality()
        +Infinity
        sage: NN.list()
        Traceback (most recent call last):
        ...
        NotImplementedError: cannot list an infinite set
        sage: NN.element_class
        <class 'sage.rings.integer.Integer'>
        sage: it = iter(NN)
        sage: [next(it), next(it), next(it), next(it), next(it)]
        [0, 1, 2, 3, 4]
        sage: x = next(it); type(x)
        <class 'sage.rings.integer.Integer'>
        sage: x.parent()
        Integer Ring
        sage: x+3
        8
        sage: NN(15)
        15
        sage: NN.first()
        0

    This checks that the different methods of `NN` return consistent
    results::

        sage: TestSuite(NN).run(verbose = True)
        running ._test_an_element() . . . pass
        running ._test_cardinality() . . . pass
        running ._test_category() . . . pass
        running ._test_construction() . . . pass
        running ._test_elements() . . .
          Running the test suite of self.an_element()
          running ._test_category() . . . pass
          running ._test_eq() . . . pass
          running ._test_new() . . . pass
          running ._test_nonzero_equal() . . . pass
          running ._test_not_implemented_methods() . . . pass
          running ._test_pickling() . . . pass
          pass
        running ._test_elements_eq_reflexive() . . . pass
        running ._test_elements_eq_symmetric() . . . pass
        running ._test_elements_eq_transitive() . . . pass
        running ._test_elements_neq() . . . pass
        running ._test_enumerated_set_contains() . . . pass
        running ._test_enumerated_set_iter_cardinality() . . . pass
        running ._test_enumerated_set_iter_list() . . . pass
        running ._test_eq() . . . pass
        running ._test_new() . . . pass
        running ._test_not_implemented_methods() . . . pass
        running ._test_pickling() . . . pass
        running ._test_some_elements() . . . pass
    """

    def __init__(self):
        """
        TESTS::

            sage: NN = InfiniteEnumeratedSets().example()
            sage: NN
            An example of an infinite enumerated set: the nonnegative integers
            sage: NN.category()
            Category of infinite enumerated sets
            sage: TestSuite(NN).run()
        """
        Parent.__init__(self, category=InfiniteEnumeratedSets())

    def _repr_(self):
        """
        TESTS::

            sage: InfiniteEnumeratedSets().example() # indirect doctest
            An example of an infinite enumerated set: the nonnegative integers
        """
        return "An example of an infinite enumerated set: the nonnegative integers"

    def __contains__(self, elt):
        """
        EXAMPLES::

            sage: NN = InfiniteEnumeratedSets().example()
            sage: 1 in NN
            True
            sage: -1 in NN
            False
        """
        return Integer(elt) >= Integer(0)

    def __iter__(self):
        """
        EXAMPLES::

            sage: NN = InfiniteEnumeratedSets().example()
            sage: g = iter(NN)
            sage: next(g), next(g), next(g), next(g)
            (0, 1, 2, 3)
        """
        i = Integer(0)
        while True:
            yield self._element_constructor_(i)
            i += 1

    def __call__(self, elt):
        """
        EXAMPLES::

            sage: NN = InfiniteEnumeratedSets().example()
            sage: NN(3)         # indirect doctest
            3
            sage: NN(3).parent()
            Integer Ring
            sage: NN(-1)
            Traceback (most recent call last):
            ...
            ValueError: Value -1 is not a nonnegative integer.
        """
        if elt in self:
            return self._element_constructor_(elt)
        raise ValueError("Value %s is not a nonnegative integer." % (elt))

    def an_element(self):
        """
        EXAMPLES::

            sage: InfiniteEnumeratedSets().example().an_element()
            42
        """
        return self._element_constructor_(Integer(42))

    def next(self, o):
        """
        EXAMPLES::

            sage: NN = InfiniteEnumeratedSets().example()
            sage: NN.next(3)
            4
        """
        return self._element_constructor_(o+1)

    def _element_constructor_(self, i):
        """
        The default implementation of _element_constructor_ assumes
        that the constructor of the element class takes the parent as
        parameter. This is not the case for ``Integer``, so we need to
        provide an implementation.


        TESTS::

            sage: NN = InfiniteEnumeratedSets().example()
            sage: x = NN(42); x
            42
            sage: type(x)
            <class 'sage.rings.integer.Integer'>
            sage: x.parent()
            Integer Ring
        """
        return self.element_class(i)

    Element = Integer


Example = NonNegativeIntegers
